import tempfile
from pathlib import Path

import pytest

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile
from genomics_data_index.storage.io.mutation.VariationFile import VariationFile
from genomics_data_index.storage.io.mutation.VcfSnpEffAnnotationParser import VcfSnpEffAnnotationParser
from genomics_data_index.test.integration import data_dir, regular_vcf_dir, variation_dir, reference_file, consensus_dir
from genomics_data_index.test.integration import extra_snippy_dir
from genomics_data_index.test.integration import reference_file_5000_snpeff, snpeff_vcf_file
from genomics_data_index.test.integration import snpeff_sample_vcfs, snpeff_sarscov2_vcfs
from genomics_data_index.test.integration.storage.io import read_vcf_df


@pytest.fixture
def snpeff_parser() -> VcfSnpEffAnnotationParser:
    return VcfSnpEffAnnotationParser()


def test_write():
    sample_vcf = data_dir / 'SampleA' / 'snps.vcf.gz'
    with tempfile.TemporaryDirectory() as out_dir:
        out_file = Path(out_dir) / 'out.vcf.gz'

        assert not out_file.exists()
        file, index = VariationFile(sample_vcf).write(out_file)
        assert out_file.exists()
        assert file == out_file
        assert index.exists()
        assert str(file) + '.csi' == str(index)

        df = read_vcf_df(out_file)
        assert 'SNP' == df[df['POS'] == 293]['TYPE'].tolist()[0]
        assert 'INDEL' == df[df['POS'] == 302]['TYPE'].tolist()[0]
        assert 'INDEL' == df[df['POS'] == 324]['TYPE'].tolist()[0]
        assert 'INDEL' == df[df['POS'] == 374]['TYPE'].tolist()[0]
        assert 'OTHER' == df[df['POS'] == 461]['TYPE'].tolist()[0]
        assert 'SNP' == df[df['POS'] == 506]['TYPE'].tolist()[0]


def test_write_2():
    sample_vcf = data_dir / 'SampleC' / 'snps.vcf.gz'
    with tempfile.TemporaryDirectory() as out_dir:
        out_file = Path(out_dir) / 'out.vcf.gz'

        assert not out_file.exists()
        file, index = VariationFile(sample_vcf).write(out_file)
        assert out_file.exists()
        assert file == out_file
        assert index.exists()
        assert str(file) + '.csi' == str(index)

        df = read_vcf_df(out_file)
        assert 'INDEL' == df[df['POS'] == 347]['TYPE'].tolist()[0]
        assert 'SNP' == df[df['POS'] == 619]['TYPE'].tolist()[0]
        assert 'OTHER' == df[df['POS'] == 1984]['TYPE'].tolist()[0]


def test_write_missing_type_tag():
    sample_vcf = regular_vcf_dir / 'SampleA.vcf.gz'
    with tempfile.TemporaryDirectory() as out_dir:
        out_file = Path(out_dir) / 'out.vcf.gz'

        assert not out_file.exists()
        VariationFile(sample_vcf).write(out_file)
        assert out_file.exists()

        df = read_vcf_df(out_file)
        assert 'SNP' == df[df['POS'] == 293]['TYPE'].tolist()[0]
        assert 'INDEL' == df[df['POS'] == 302]['TYPE'].tolist()[0]
        assert 'INDEL' == df[df['POS'] == 324]['TYPE'].tolist()[0]
        assert 'INDEL' == df[df['POS'] == 374]['TYPE'].tolist()[0]
        assert 'OTHER' == df[df['POS'] == 461]['TYPE'].tolist()[0]
        assert 'SNP' == df[df['POS'] == 506]['TYPE'].tolist()[0]


def test_write_2_missing_type_tag():
    sample_vcf = regular_vcf_dir / 'SampleC.vcf.gz'
    with tempfile.TemporaryDirectory() as out_dir:
        out_file = Path(out_dir) / 'out.vcf.gz'

        assert not out_file.exists()
        VariationFile(sample_vcf).write(out_file)
        assert out_file.exists()

        df = read_vcf_df(out_file)
        assert 'INDEL' == df[df['POS'] == 347]['TYPE'].tolist()[0]
        assert 'SNP' == df[df['POS'] == 619]['TYPE'].tolist()[0]
        assert 'OTHER' == df[df['POS'] == 1984]['TYPE'].tolist()[0]


def test_write_bcf():
    sample_vcf = data_dir / 'SampleA' / 'snps.vcf.gz'
    with tempfile.TemporaryDirectory() as out_dir:
        out_file = Path(out_dir) / 'out.bcf'

        assert not out_file.exists()
        VariationFile(sample_vcf).write(out_file)
        assert out_file.exists()


def test_consensus_no_mask():
    sample_bcf = variation_dir / 'SampleA.bcf'

    expected_consensus_file = consensus_dir / 'SampleA-consensus-nomask.fasta.gz'
    name, expected_consensus_records = SequenceFile(expected_consensus_file).parse_sequence_file()
    expected_consensus_record = expected_consensus_records[0]

    seq_records = VariationFile(sample_bcf).consensus(reference_file=reference_file)

    assert 1 == len(seq_records)
    actual_seq_record = seq_records[0]
    assert 5180 == len(actual_seq_record)
    assert expected_consensus_record.id == actual_seq_record.id
    assert expected_consensus_record.seq == actual_seq_record.seq


def test_consensus_empty_mask():
    sample_bcf = variation_dir / 'SampleA.bcf'
    empty_mask = MaskedGenomicRegions.empty_mask()

    expected_consensus_file = consensus_dir / 'SampleA-consensus-nomask.fasta.gz'
    name, expected_consensus_records = SequenceFile(expected_consensus_file).parse_sequence_file()
    expected_consensus_record = expected_consensus_records[0]

    with tempfile.TemporaryDirectory() as out_dir:
        mask_file = Path(out_dir) / 'mask.bed.gz'
        empty_mask.write(mask_file)

        seq_records = VariationFile(sample_bcf).consensus(reference_file=reference_file,
                                                          mask_file=mask_file)
        assert 1 == len(seq_records)
        actual_seq_record = seq_records[0]
        assert 5180 == len(actual_seq_record)
        assert expected_consensus_record.id == actual_seq_record.id
        assert expected_consensus_record.seq == actual_seq_record.seq


def test_consensus_mask():
    sample_bcf = variation_dir / 'SampleA.bcf'
    sample_mask_file = variation_dir / 'SampleA.bed.gz'

    expected_consensus_file = consensus_dir / 'SampleA-consensus-withmask.fasta.gz'
    name, expected_consensus_records = SequenceFile(expected_consensus_file).parse_sequence_file()
    expected_consensus_record = expected_consensus_records[0]

    seq_records = VariationFile(sample_bcf).consensus(reference_file=reference_file,
                                                      mask_file=sample_mask_file)
    assert 1 == len(seq_records)
    actual_seq_record = seq_records[0]
    assert 5180 == len(actual_seq_record)
    assert expected_consensus_record.id == actual_seq_record.id
    assert expected_consensus_record.seq == actual_seq_record.seq


def test_consensus_mask_over_mutation():
    sample_bcf = variation_dir / 'SampleA.bcf'
    sample_mask_file = variation_dir / 'SampleA-mask-over-mutation.bed.gz'

    expected_consensus_file = consensus_dir / 'SampleA-consensus-withmask-over-mutation.fasta.gz'
    name, expected_consensus_records = SequenceFile(expected_consensus_file).parse_sequence_file()
    expected_consensus_record = expected_consensus_records[0]

    seq_records = VariationFile(sample_bcf).consensus(reference_file=reference_file,
                                                      mask_file=sample_mask_file)
    assert 1 == len(seq_records)
    actual_seq_record = seq_records[0]
    assert 5180 == len(actual_seq_record)
    assert expected_consensus_record.id == actual_seq_record.id
    assert expected_consensus_record.seq == actual_seq_record.seq


def test_union_all_files():
    # List like this to guarantee a specific order
    variant_files = [
        data_dir / 'SampleA' / 'snps.vcf.gz',
        data_dir / 'SampleB' / 'snps.vcf.gz',
        data_dir / 'SampleC' / 'snps.vcf.gz'
    ]
    union_df = VariationFile.union_all_files(variant_files, include_expression='TYPE="SNP"')

    assert 60 == len(union_df)
    assert ['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'] == union_df.columns.tolist()
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 190)]['COUNT'].values[0]
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 5061)]['COUNT'].values[0]
    assert 2 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 4975)]['COUNT'].values[0]
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 2076)]['COUNT'].values[0]


def test_union_one_file():
    sample_bcf = variation_dir / 'SampleA.bcf'
    union_df = VariationFile.union_all_files([sample_bcf], include_expression='TYPE="SNP"')

    assert 26 == len(union_df)
    assert ['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'] == union_df.columns.tolist()
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 293)]['COUNT'].values[0]
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 4929)]['COUNT'].values[0]


def test_union_batch_size_1():
    # List like this to guarantee a specific order
    variant_files = [
        data_dir / 'SampleA' / 'snps.vcf.gz',
        data_dir / 'SampleB' / 'snps.vcf.gz',
        data_dir / 'SampleC' / 'snps.vcf.gz'
    ]
    union_df = VariationFile.union_all_files(variant_files, include_expression='TYPE="SNP"', batch_size=1)

    assert 60 == len(union_df)
    assert ['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'] == union_df.columns.tolist()
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 190)]['COUNT'].values[0]
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 5061)]['COUNT'].values[0]
    assert 2 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 4975)]['COUNT'].values[0]
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 2076)]['COUNT'].values[0]


def test_union_batch_size_2_all_data():
    # List like this to guarantee a specific order
    variant_files = [
        data_dir / 'SampleA' / 'snps.vcf.gz',
        data_dir / 'SampleB' / 'snps.vcf.gz',
        data_dir / 'SampleC' / 'snps.vcf.gz'
    ]
    union_df = VariationFile.union_all_files(variant_files, batch_size=2)

    assert 112 == len(union_df)
    assert ['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'] == union_df.columns.tolist()
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 190)]['COUNT'].values[0]
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 5061)]['COUNT'].values[0]
    assert 2 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 4975)]['COUNT'].values[0]
    assert 1 == union_df[(union_df['CHROM'] == 'reference') & (union_df['POS'] == 2076)]['COUNT'].values[0]


def test_union_many_files_batch_size_2_more_data():
    # List like this to guarantee a specific order
    variant_files = [
        data_dir / 'SampleA' / 'snps.vcf.gz',
        data_dir / 'SampleB' / 'snps.vcf.gz',
        data_dir / 'SampleC' / 'snps.vcf.gz',
        extra_snippy_dir / 'SampleB2.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB3.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB4.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB5.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB5-different-allele.fill-tags.vcf.gz',
    ]
    union_df = VariationFile.union_all_files(variant_files, batch_size=2)

    assert 115 == len(union_df)
    assert ['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'] == union_df.columns.tolist()
    assert 6 == union_df[union_df['ID'] == 'reference:190:A:G']['COUNT'].values[0]
    assert 6 == union_df[union_df['ID'] == 'reference:5061:G:A']['COUNT'].values[0]

    assert 6 == union_df[union_df['ID'] == 'reference:4975:T:C']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:4975:T:CAT']['COUNT'].values[0]

    assert 1 == union_df[union_df['ID'] == 'reference:2076:A:T']['COUNT'].values[0]

    assert 2 == union_df[union_df['ID'] == 'reference:1483:AAAGAGGGGCTGCTGGAGCCG:A']['COUNT'].values[0]
    assert 3 == union_df[union_df['ID'] == 'reference:1640:C:A']['COUNT'].values[0]

    assert 6 == union_df[union_df['ID'] == 'reference:4693:C:CGA']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:4693:C:G']['COUNT'].values[0]

    assert 3 == union_df[union_df['ID'] == 'reference:883:CACATG:C']['COUNT'].values[0]

    assert 5 == union_df[union_df['ID'] == 'reference:349:AAGT:A']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:349:AAGT:T']['COUNT'].values[0]


def test_union_many_files_batch_size_2_with_empty_vcf():
    # List like this to guarantee a specific order
    variant_files = [
        data_dir / 'SampleA' / 'snps.vcf.gz',
        data_dir / 'SampleB' / 'snps.vcf.gz',
        data_dir / 'SampleC' / 'snps.vcf.gz',
        extra_snippy_dir / 'SampleB2.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB3.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB4.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB5.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB5-different-allele.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB-empty.snps.fill-tags.vcf.gz'
    ]
    union_df = VariationFile.union_all_files(variant_files, batch_size=2)
    print(union_df)

    assert 115 == len(union_df)
    assert ['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'] == union_df.columns.tolist()
    assert 6 == union_df[union_df['ID'] == 'reference:190:A:G']['COUNT'].values[0]
    assert 6 == union_df[union_df['ID'] == 'reference:5061:G:A']['COUNT'].values[0]

    assert 6 == union_df[union_df['ID'] == 'reference:4975:T:C']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:4975:T:CAT']['COUNT'].values[0]

    assert 1 == union_df[union_df['ID'] == 'reference:2076:A:T']['COUNT'].values[0]

    assert 2 == union_df[union_df['ID'] == 'reference:1483:AAAGAGGGGCTGCTGGAGCCG:A']['COUNT'].values[0]
    assert 3 == union_df[union_df['ID'] == 'reference:1640:C:A']['COUNT'].values[0]

    assert 6 == union_df[union_df['ID'] == 'reference:4693:C:CGA']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:4693:C:G']['COUNT'].values[0]

    assert 3 == union_df[union_df['ID'] == 'reference:883:CACATG:C']['COUNT'].values[0]

    assert 5 == union_df[union_df['ID'] == 'reference:349:AAGT:A']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:349:AAGT:T']['COUNT'].values[0]


def test_union_many_files_batch_size_odd_cores_3():
    # List like this to guarantee a specific order
    variant_files = [
        data_dir / 'SampleA' / 'snps.vcf.gz',
        data_dir / 'SampleB' / 'snps.vcf.gz',
        data_dir / 'SampleC' / 'snps.vcf.gz',
        extra_snippy_dir / 'SampleB2.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB3.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB4.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB5.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB5-different-allele.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB-empty.snps.fill-tags.vcf.gz'
    ]
    union_df = VariationFile.union_all_files(variant_files, ncores=3, batch_size=3)
    print(union_df)

    assert 115 == len(union_df)
    assert ['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'] == union_df.columns.tolist()
    assert 6 == union_df[union_df['ID'] == 'reference:190:A:G']['COUNT'].values[0]
    assert 6 == union_df[union_df['ID'] == 'reference:5061:G:A']['COUNT'].values[0]

    assert 6 == union_df[union_df['ID'] == 'reference:4975:T:C']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:4975:T:CAT']['COUNT'].values[0]

    assert 1 == union_df[union_df['ID'] == 'reference:2076:A:T']['COUNT'].values[0]

    assert 2 == union_df[union_df['ID'] == 'reference:1483:AAAGAGGGGCTGCTGGAGCCG:A']['COUNT'].values[0]
    assert 3 == union_df[union_df['ID'] == 'reference:1640:C:A']['COUNT'].values[0]

    assert 6 == union_df[union_df['ID'] == 'reference:4693:C:CGA']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:4693:C:G']['COUNT'].values[0]

    assert 3 == union_df[union_df['ID'] == 'reference:883:CACATG:C']['COUNT'].values[0]

    assert 5 == union_df[union_df['ID'] == 'reference:349:AAGT:A']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:349:AAGT:T']['COUNT'].values[0]


def test_union_many_files_ambiguous():
    # List like this to guarantee a specific order
    variant_files = [
        data_dir / 'SampleA' / 'snps.vcf.gz',
        data_dir / 'SampleB' / 'snps.vcf.gz',
        data_dir / 'SampleC' / 'snps.vcf.gz',
        extra_snippy_dir / 'SampleB2.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB3.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB4.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB5.snps.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB5-different-allele.fill-tags.vcf.gz',
        extra_snippy_dir / 'SampleB5-different-allele-ambiguous.vcf.gz',
    ]
    union_df = VariationFile.union_all_files(variant_files)
    print(union_df)

    assert 119 == len(union_df)
    assert ['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'] == union_df.columns.tolist()
    assert 6 == union_df[union_df['ID'] == 'reference:190:A:G']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:190:A:N']['COUNT'].values[0]

    assert 7 == union_df[union_df['ID'] == 'reference:5061:G:A']['COUNT'].values[0]
    assert 2 == union_df[union_df['ID'] == 'reference:1483:AAAGAGGGGCTGCTGGAGCCG:A']['COUNT'].values[0]

    assert 5 == union_df[union_df['ID'] == 'reference:349:AAGT:A']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:349:AAGT:T']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:349:ANGT:T']['COUNT'].values[0]

    assert 6 == union_df[union_df['ID'] == 'reference:4693:C:CGA']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:4693:C:G']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:4693:N:G']['COUNT'].values[0]

    assert 6 == union_df[union_df['ID'] == 'reference:4975:T:C']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:4975:T:CAT']['COUNT'].values[0]
    assert 1 == union_df[union_df['ID'] == 'reference:4975:T:CNT']['COUNT'].values[0]


def test_union_many_files_batch_size_2_single_empty_vcf():
    # List like this to guarantee a specific order
    variant_files = [
        extra_snippy_dir / 'SampleB-empty.snps.fill-tags.vcf.gz'
    ]
    union_df = VariationFile.union_all_files(variant_files, batch_size=2)
    print(union_df)

    assert 0 == len(union_df)
    assert ['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'] == union_df.columns.tolist()


def test_read_features(snpeff_parser):
    vcf_file = data_dir / 'SampleA' / 'snps.vcf.gz'
    df = VariationFile(vcf_file).read_features('SampleA', snpeff_parser=snpeff_parser)

    assert 46 == len(df), 'Data fram has incorrect length'

    assert {'snps.vcf.gz'} == set(df['FILE'].tolist()), 'Incorrect filename'
    assert {'SampleA'} == set(df['SAMPLE'].tolist()), 'Incorrect sample name'

    v = df[df['POS'] == 461]
    assert 'AAAT' == v['REF'].values[0], 'Incorrect reference'
    assert 'G' == v['ALT'].values[0], 'Incorrect alt'

    v = df[df['POS'] == 1048]
    assert 'C' == v['REF'].values[0], 'Incorrect reference'
    assert 'G' == v['ALT'].values[0], 'Incorrect alt'

    v = df[df['POS'] == 1253]
    assert 'T' == v['REF'].values[0], 'Incorrect reference'
    assert 'TAA' == v['ALT'].values[0], 'Incorrect alt'

    v = df[df['POS'] == 3656]
    assert 'CATT' == v['REF'].values[0], 'Incorrect reference'
    assert 'C' == v['ALT'].values[0], 'Incorrect alt'


def test_read_features_snpeff(snpeff_parser):
    sample_10_014 = VariationFile(
        snpeff_sample_vcfs['SH10-014']).read_features('SH10-014', snpeff_parser=snpeff_parser).sort_values('POS')
    sample_14_001 = VariationFile(
        snpeff_sample_vcfs['SH14-001']).read_features('SH14-001', snpeff_parser=snpeff_parser).sort_values('POS')
    sample_14_014 = VariationFile(
        snpeff_sample_vcfs['SH14-014']).read_features('SH14-014', snpeff_parser=snpeff_parser).sort_values('POS')

    assert 139 == len(sample_10_014)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(sample_10_014.columns)

    # snv/snp
    sample_10_014_varA = sample_10_014[sample_10_014['POS'] == 140658]
    assert 1 == len(sample_10_014_varA)
    assert ['SH10-014', 'NC_011083', 140658, 'C', 'A', 'snp', 'SH10-014.vcf.gz', 'NC_011083:140658:C:A',
            'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu'] == sample_10_014_varA[
               sample_10_014_varA['ANN.Annotation'] == 'missense_variant'].iloc[0].tolist()

    # del
    sample_10_014_varB = sample_10_014[sample_10_014['POS'] == 1125996]
    assert 1 == len(sample_10_014_varB)
    assert ['SH10-014', 'NC_011083', 1125996, 'CG', 'C', 'del', 'SH10-014.vcf.gz', 'NC_011083:1125996:CG:C',
            'C', 'frameshift_variant', 'HIGH', 'SEHA_RS05995', 'SEHA_RS05995', 'transcript', 'protein_coding',
            'c.418delG', 'p.Glu140fs'] == sample_10_014_varB[
               sample_10_014_varB['ANN.Annotation'] == 'frameshift_variant'].iloc[0].tolist()

    # ins
    sample_10_014_varC = sample_10_014[sample_10_014['POS'] == 1246085]
    assert 1 == len(sample_10_014_varC)
    assert ['SH10-014', 'NC_011083', 1246085, 'C', 'CG', 'ins', 'SH10-014.vcf.gz', 'NC_011083:1246085:C:CG',
            'CG', 'frameshift_variant', 'HIGH', 'mdtG', 'SEHA_RS06605', 'transcript', 'protein_coding',
            'c.722dupC', 'p.Leu242fs'] == sample_10_014_varC[
               sample_10_014_varC['ANN.Annotation'] == 'frameshift_variant'].iloc[0].tolist()

    # complex
    sample_10_014_varD = sample_10_014[sample_10_014['POS'] == 3535121]
    assert 1 == len(sample_10_014_varD)
    assert ['SH10-014', 'NC_011083', 3535121, 'CGCGA', 'TGTGG', 'complex', 'SH10-014.vcf.gz',
            'NC_011083:3535121:CGCGA:TGTGG',
            'TGTGG', 'missense_variant', 'MODERATE', 'oadA', 'SEHA_RS17780', 'transcript', 'protein_coding',
            'c.1119_1123delTCGCGinsCCACA', 'p.ArgAla374HisThr'] == sample_10_014_varD[
               sample_10_014_varD['ANN.Annotation'] == 'missense_variant'].iloc[0].tolist()

    assert 115 == len(sample_14_001)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(sample_14_001.columns)
    sample_14_001_var = sample_14_001[sample_14_001['POS'] == 140658]
    assert 1 == len(sample_14_001_var)
    assert ['SH14-001', 'NC_011083', 140658, 'C', 'A', 'snp', 'SH14-001.vcf.gz', 'NC_011083:140658:C:A',
            'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu'] == sample_14_001_var[
               sample_14_001_var['ANN.Annotation'] == 'missense_variant'].iloc[0].tolist()

    assert 107 == len(sample_14_014)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(sample_14_014.columns)
    sample_14_014_var = sample_14_014[sample_14_014['POS'] == 298472]
    assert 1 == len(sample_14_014_var)
    assert ['SH14-014', 'NC_011083', 298472, 'A', 'C', 'snp', 'SH14-014.vcf.gz', 'NC_011083:298472:A:C',
            'C', 'intergenic_region', 'MODIFIER', 'SEHA_RS01880-SEHA_RS01885', 'SEHA_RS01880-SEHA_RS01885',
            'intergenic_region', 'n.298472A>C'] == sample_14_014_var[
               sample_14_014_var['ANN.Annotation'] == 'intergenic_region'].drop(
        ['ANN.Transcript_BioType', 'ANN.HGVS.p'], axis='columns').iloc[0].tolist()
    assert {True} == set(sample_14_014_var[sample_14_014_var['ANN.Annotation'] == 'intergenic_region'] \
                             [['ANN.Transcript_BioType', 'ANN.HGVS.p']].iloc[0].isna().tolist())


def test_read_features_snpeff_sars_cov_2(snpeff_parser):
    sample_sarscov2_1 = VariationFile(
        snpeff_sarscov2_vcfs['USA/CA-CDPH-3000143037/2021']
    ).read_features('USA/CA-CDPH-3000143037/2021', snpeff_parser=snpeff_parser).sort_values('POS')

    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(sample_sarscov2_1.columns)
    assert 31 == len(sample_sarscov2_1)

    sample_sarscov2_1['ANN.Annotation'] = sample_sarscov2_1['ANN.Annotation'].astype(str)
    sample_sarscov2_1['ANN.Annotation_Impact'] = sample_sarscov2_1['ANN.Annotation_Impact'].astype(str)

    # ORF1ab (ORF1a region)
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 3948]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'USA__CA-CDPH-3000143037__2021.vcf.gz',
            'NC_045512.2:3948:A:G', 'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript',
            'protein_coding', 'c.3683A>G', 'p.D1228G'] == sample_sarscov2_1_var.iloc[0].tolist()

    # ORF1ab (ORF1a region)
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 3037]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 3037, 'C', 'T', 'SNP', 'USA__CA-CDPH-3000143037__2021.vcf.gz',
            'NC_045512.2:3037:C:T', 'T', 'synonymous_variant', 'LOW', 'ORF1ab', 'GU280_gp01', 'transcript',
            'protein_coding', 'c.2772C>T', 'p.F924F'] == sample_sarscov2_1_var.iloc[0].tolist()

    # ORF1ab (ORF1b region)
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 19220]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 19220, 'C', 'T', 'SNP',
            'USA__CA-CDPH-3000143037__2021.vcf.gz',
            'NC_045512.2:19220:C:T', 'T', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript',
            'protein_coding', 'c.18956C>T', 'p.A6319V'] == sample_sarscov2_1_var.iloc[0].tolist()

    # S
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 22917]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 22917, 'T', 'G', 'SNP',
            'USA__CA-CDPH-3000143037__2021.vcf.gz',
            'NC_045512.2:22917:T:G', 'G', 'missense_variant', 'MODERATE', 'S', 'GU280_gp02', 'transcript',
            'protein_coding', 'c.1355T>G', 'p.L452R'] == sample_sarscov2_1_var.iloc[0].tolist()

    # ORF7b
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 27874]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 27874, 'C', 'T', 'SNP',
            'USA__CA-CDPH-3000143037__2021.vcf.gz',
            'NC_045512.2:27874:C:T', 'T', 'missense_variant', 'MODERATE', 'ORF7b', 'GU280_gp08', 'transcript',
            'protein_coding', 'c.119C>T', 'p.T40I'] == sample_sarscov2_1_var.iloc[0].tolist()

    # intergenic
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 210].fillna('<NA>')
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 210, 'G', 'T', 'SNP', 'USA__CA-CDPH-3000143037__2021.vcf.gz',
            'NC_045512.2:210:G:T', 'T', 'intergenic_region', 'MODIFIER', 'CHR_START-ORF1ab', 'CHR_START-GU280_gp01',
            'intergenic_region', '<NA>', 'n.210G>T', '<NA>'] == sample_sarscov2_1_var.iloc[0].tolist()


def test_read_features_sars_cov_2_no_snpeff_annotation(snpeff_parser):
    sample_sarscov2_1 = VariationFile(
        snpeff_sarscov2_vcfs['USA/CA-CDPH-3000143037/2021.noann']
    ).read_features('USA/CA-CDPH-3000143037/2021', snpeff_parser=snpeff_parser).sort_values('POS')

    ann_columns = ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                   'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(sample_sarscov2_1.columns)
    assert 31 == len(sample_sarscov2_1)

    # ORF1ab (ORF1a region)
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 3948]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 3948, 'A', 'G', 'SNP',
            'USA__CA-CDPH-3000143037__2021.noann.vcf.gz', 'NC_045512.2:3948:A:G'] == sample_sarscov2_1_var.drop(
        ann_columns, axis='columns').iloc[0].tolist()
    assert {True} == set(sample_sarscov2_1_var[ann_columns].iloc[0].isna().tolist())

    # ORF1ab (ORF1a region)
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 3037]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 3037, 'C', 'T', 'SNP',
            'USA__CA-CDPH-3000143037__2021.noann.vcf.gz', 'NC_045512.2:3037:C:T'] == sample_sarscov2_1_var.drop(
        ann_columns, axis='columns').iloc[0].tolist()
    assert {True} == set(sample_sarscov2_1_var[ann_columns].iloc[0].isna().tolist())

    # ORF1ab (ORF1b region)
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 19220]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 19220, 'C', 'T', 'SNP',
            'USA__CA-CDPH-3000143037__2021.noann.vcf.gz', 'NC_045512.2:19220:C:T'] == sample_sarscov2_1_var.drop(
        ann_columns, axis='columns').iloc[0].tolist()
    assert {True} == set(sample_sarscov2_1_var[ann_columns].iloc[0].isna().tolist())

    # S
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 22917]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 22917, 'T', 'G', 'SNP',
            'USA__CA-CDPH-3000143037__2021.noann.vcf.gz', 'NC_045512.2:22917:T:G'] == sample_sarscov2_1_var.drop(
        ann_columns, axis='columns').iloc[0].tolist()
    assert {True} == set(sample_sarscov2_1_var[ann_columns].iloc[0].isna().tolist())

    # ORF7b
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 27874]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 27874, 'C', 'T', 'SNP',
            'USA__CA-CDPH-3000143037__2021.noann.vcf.gz', 'NC_045512.2:27874:C:T'] == sample_sarscov2_1_var.drop(
        ann_columns, axis='columns').iloc[0].tolist()
    assert {True} == set(sample_sarscov2_1_var[ann_columns].iloc[0].isna().tolist())

    # intergenic
    sample_sarscov2_1_var = sample_sarscov2_1[sample_sarscov2_1['POS'] == 210]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 210, 'G', 'T', 'SNP',
            'USA__CA-CDPH-3000143037__2021.noann.vcf.gz', 'NC_045512.2:210:G:T'] == sample_sarscov2_1_var.drop(
        ann_columns, axis='columns').iloc[0].tolist()
    assert {True} == set(sample_sarscov2_1_var[ann_columns].iloc[0].isna().tolist())


def test_annotate(snpeff_parser):
    with tempfile.TemporaryDirectory() as out_dir:
        database_dir = Path(out_dir)
        output_vcf_file = database_dir / 'output.vcf.gz'
        variation_file = VariationFile(snpeff_vcf_file)

        snpeff_database = SequenceFile(reference_file_5000_snpeff).create_snpeff_database(database_dir)
        annotated_variation_file = variation_file.annotate(snpeff_database=snpeff_database,
                                                           annotated_vcf=output_vcf_file)

        assert output_vcf_file == annotated_variation_file.file

        # Verify VCF annotation contents
        vcf_annotation_df = annotated_variation_file.read_features('SampleA', snpeff_parser=snpeff_parser).sort_values(
            'POS')
        assert 2 == len(vcf_annotation_df)
        assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
                'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
            vcf_annotation_df.columns)
        assert ['NC_011083.1:195:C:G', 'NC_011083.1:207:C:G'] == vcf_annotation_df['VARIANT_ID'].tolist()
        assert ['SNP', 'SNP'] == vcf_annotation_df['TYPE'].tolist()
        assert ['missense_variant', 'synonymous_variant'] == vcf_annotation_df['ANN.Annotation'].tolist()
        assert ['SEHA_RS00560', 'SEHA_RS00560'] == vcf_annotation_df['ANN.Gene_ID'].tolist()
        assert ['thrL', 'thrL'] == vcf_annotation_df['ANN.Gene_Name'].tolist()
        assert ['c.6C>G', 'c.18C>G'] == vcf_annotation_df['ANN.HGVS.c'].tolist()
        assert ['p.N2K', 'p.T6T'] == vcf_annotation_df['ANN.HGVS.p'].tolist()

        # Original file should still exist and be unannotated
        vcf_no_annotation_df = variation_file.read_features('SampleA', snpeff_parser=snpeff_parser).sort_values('POS')
        assert 2 == len(vcf_no_annotation_df)
        assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
                'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
            vcf_no_annotation_df.columns)
        assert ['NC_011083.1:195:C:G', 'NC_011083.1:207:C:G'] == vcf_no_annotation_df['VARIANT_ID'].tolist()
        assert ['SNP', 'SNP'] == vcf_no_annotation_df['TYPE'].tolist()
        assert all(vcf_no_annotation_df['ANN.Annotation'].isna())
        assert all(vcf_no_annotation_df['ANN.Gene_ID'].isna())
        assert all(vcf_no_annotation_df['ANN.HGVS.c'].isna())
        assert all(vcf_no_annotation_df['ANN.HGVS.p'].isna())



import gzip
from os import path
from pathlib import Path

import pytest
from Bio import SeqIO

from genomics_data_index.api.query.kind.isa.typing.ExperimentalSARSCov2ConstellationsTyper import \
    ExperimentalSARSCov2ConstellationsTyper

root_data_dir = Path(path.dirname(__file__)) / '..' / '..' / '..' / 'data'
sequence_path = root_data_dir / 'NC_045512.2.gb.gz'


@pytest.fixture
def sequence_record():
    with gzip.open(sequence_path, mode='rt') as fh:
        sequences = list(SeqIO.parse(fh, 'genbank'))
        return sequences[0]


def test_mutation_to_identifier(sequence_record):
    typer = ExperimentalSARSCov2ConstellationsTyper(constellation_files=[],
                                                    sequence=sequence_record)

    assert 'hgvs_gn:NC_045512.2:S:p.H69_V70del' == typer.mutation_to_identifier('s:HV69-')
    assert 'hgvs_gn:NC_045512.2:ORF1ab:p.S3675_F3677del' == typer.mutation_to_identifier('1ab:SGF3675-')
    assert 'NC_045512.2:5986:C:T' == typer.mutation_to_identifier('nuc:C5986T')
    assert 'hgvs_gn:NC_045512.2:ORF8:p.Q27*' == typer.mutation_to_identifier('8:Q27*')
    assert 'hgvs_gn:NC_045512.2:S:p.Y144del' == typer.mutation_to_identifier('s:Y144-')

    assert 'NC_045512.2:4:AA:A' == typer.mutation_to_identifier('del:4:1')
    assert 'NC_045512.2:4:AAA:A' == typer.mutation_to_identifier('del:4:2')
    assert 'NC_045512.2:4:AAAG:A' == typer.mutation_to_identifier('del:4:3')
    assert 'NC_045512.2:21:CAGGTAACAAA:C' == typer.mutation_to_identifier('del:21:10')



from typing import List, Any, cast, Optional

import pandas as pd

from genomics_data_index.api.query.features.FeaturesFromIndexComparator import FeatureSamplesSingleCategorySummarizer
from genomics_data_index.api.query.features.FeaturesFromIndexComparator import FeatureSamplesSummarizer
from genomics_data_index.api.query.features.FeaturesFromIndexComparator import FeaturesFromIndexComparator
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI
from genomics_data_index.storage.model.db import FeatureSamples
from genomics_data_index.storage.model.db import NucleotideVariantsSamples
from genomics_data_index.storage.service.VariationService import VariationService


class MutationFeaturesFromIndexComparator(FeaturesFromIndexComparator):

    def __init__(self, connection: DataIndexConnection, ignore_annotations: bool = False,
                 include_present: bool = True, include_unknown: bool = False,
                 include_unknown_samples: bool = True,
                 include_unknown_no_present_samples=False,
                 mutation_type: str = 'all', id_type: str = 'spdi_ref'):
        super().__init__(connection=connection, include_unknown_samples=include_unknown_samples,
                         include_unknown_no_present_samples=include_unknown_no_present_samples)
        self._ignore_annotations = ignore_annotations
        self._mutation_type = mutation_type
        self._include_present = include_present
        self._include_unknown = include_unknown
        self._id_type = id_type

    @property
    def summary_columns(self) -> List[str]:
        return self.feature_id_columns + FeatureSamplesSingleCategorySummarizer.SUMMARY_NAMES

    @property
    def feature_id_columns(self) -> List[str]:
        return ['Sequence', 'Position', 'Deletion', 'Insertion', 'Type']

    @property
    def index_name(self) -> str:
        return 'Mutation'

    def _create_feature_sample_count_row(self, feature_id: str, feature: FeatureSamples,
                                         feature_samples: SampleSet,
                                         feature_samples_unknown: Optional[SampleSet],
                                         total: int,
                                         feature_samples_summarizer: FeatureSamplesSummarizer) -> List[Any]:
        if isinstance(feature, NucleotideVariantsSamples):
            feature = cast(NucleotideVariantsSamples, feature)
            summary_data = feature_samples_summarizer.summary_data(samples=feature_samples,
                                                                   unknown_samples=feature_samples_unknown,
                                                                   total=total)
            query_feature_id = QueryFeatureMutationSPDI(
                feature_id)  # Use this since db deletion is an int (not sequence)
            return [feature_id, feature.sequence, feature.position,
                    query_feature_id.deletion, feature.insertion, feature.var_type] + summary_data
        else:
            raise Exception(f'feature={feature} is not of type {NucleotideVariantsSamples.__name__}')

    def _join_additional_columns(self, features_df: pd.DataFrame) -> pd.DataFrame:
        if not self._ignore_annotations:
            return self._connection.variation_service.append_mutation_annotations(features_df)
        else:
            return features_df

    def _do_summary(self, sample_set: SampleSet, feature_samples_summarizer: FeatureSamplesSummarizer) -> pd.DataFrame:
        variation_service: VariationService = self._connection.variation_service
        present_features = variation_service.get_features(include_present=self._include_present,
                                                          include_unknown=self._include_unknown,
                                                          id_type=self._id_type)

        features_df = self._create_summary_df(present_features, present_samples=sample_set,
                                              feature_samples_summarizer=feature_samples_summarizer)

        return self._join_additional_columns(features_df)



import ete3
import numpy as np
import pandas as pd
import scipy.cluster.hierarchy as sch
import skbio
import skbio.tree
from ete3 import Tree
from scipy.spatial.distance import jensenshannon
from scipy.spatial.distance import squareform


class SampleCategoriesClustererExperimental:
    """
    This class contains experimental methods for clustering sample categories (e.g., lineage or other
    metadata) by the features found in these categories. This is called experimental since I have
    not had time to properly test these out yet. Eventually they may make it into the rest of the API.
    """

    def __init__(self):
        pass

    def cluster_features_comparison(self, comparison_df: pd.DataFrame) -> Tree:
        """
        Clusters vectors of feature proportions in particular categories by their Jensen-Shannon distance.
        That is, takes as input a dataframe generated from "SamplesQuery.features_comparison(..., unit='proportion')"
        (which compares the proportion of genomes with a particular feature in some category such as a linage). Produces a
        pair-wise distance matrix from the distances between these feature vectors and clusters this distance matrix
        (using single-linkage clustering).
        :param comparison_df: The dataframe produced as output from "SamplesQuery.features_comparison(..., unit='proportion')"
        :return: An ete3.ClusterTree which stores the hierarchical cluster of each of the sample categories.
        """
        comparison_df = comparison_df.filter(like='_proportion')
        categories = [c[:-len('_proportion')] for c in comparison_df]

        # Create 2d-array of distances
        distances = []
        for i1, c1 in enumerate(categories):
            row = []
            for i2, c2 in enumerate(categories):
                vec1 = comparison_df[f'{c1}_proportion'].to_numpy()
                vec2 = comparison_df[f'{c2}_proportion'].to_numpy()
                row.append(jensenshannon(vec1, vec2))
            distances.append(row)

        distance_matrix = np.array(distances)
        compressed_matrix = squareform(distance_matrix)
        linkage_matrix = sch.linkage(compressed_matrix, method='single')
        skb_tree = skbio.tree.TreeNode.from_linkage_matrix(linkage_matrix, id_list=categories)
        skb_newick = str(skb_tree).strip().replace("'", "")

        return ete3.ClusterNode(newick=skb_newick)



import shutil
import tempfile
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from genomics_data_index.api.query.GenomicsDataIndex import GenomicsDataIndex
from genomics_data_index.configuration.Project import Project
from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.io.processor.SerialSampleFilesProcessor import SerialSampleFilesProcessor
from genomics_data_index.storage.service import EntityExistsError
from genomics_data_index.test.integration import sample_dirs, reference_file


def test_get_reference_tree(loaded_database_genomic_data_store_with_tree: GenomicsDataIndex):
    tree = loaded_database_genomic_data_store_with_tree.reference_tree('genome')
    assert tree is not None

    with pytest.raises(EntityExistsError) as execinfo:
        loaded_database_genomic_data_store_with_tree.reference_tree('invalid_reference')

    assert 'No reference genome with name=[invalid_reference]' in str(execinfo.value)


def test_mlst_schemes(loaded_database_genomic_data_store: GenomicsDataIndex):
    mlst_schemes = loaded_database_genomic_data_store.mlst_schemes()
    assert 3 == len(mlst_schemes)
    assert isinstance(mlst_schemes, list)
    assert {'lmonocytogenes', 'ecoli', 'campylobacter'} == set(mlst_schemes)


def test_summaries_loaded_data(loaded_database_genomic_data_store: GenomicsDataIndex):
    gds = loaded_database_genomic_data_store

    # Samples
    assert 9 == gds.count_samples()
    assert {'2014C-3598', '2014C-3599', '2014D-0067', '2014D-0068',
            'CFSAN002349', 'CFSAN023463',
            'SampleA', 'SampleB', 'SampleC'} == set(gds.sample_names())

    # References
    assert 1 == gds.count_references()
    assert ['genome'] == gds.reference_names()

    assert 112 == gds.count_mutations('genome')
    assert 112 + 440 == gds.count_mutations('genome', include_unknown=True)

    # Mutations ignore unknown spdi
    ms = gds.mutations_summary('genome', id_type='spdi', ignore_annotations=True,
                               include_unknown_samples=False)
    assert 112 == len(ms)
    assert 'Mutation' == ms.index.name
    assert ['Sequence', 'Position', 'Deletion', 'Insertion', 'Type',
            'Count', 'Unknown Count', 'Present and Unknown Count', 'Total',
            'Percent', 'Unknown Percent', 'Present and Unknown Percent'] == list(ms.columns)

    ## Convert percent to int to make it easier to compare in assert statements
    ms['Percent'] = ms['Percent'].astype(int)
    ms['Unknown Count'] = ms['Unknown Count'].fillna(-1).astype(int)
    ms['Present and Unknown Count'] = ms['Present and Unknown Count'].fillna(-1).astype(int)
    ms['Unknown Percent'] = ms['Unknown Percent'].fillna(-1).astype(int)
    ms['Present and Unknown Percent'] = ms['Present and Unknown Percent'].fillna(-1).astype(int)

    assert ['reference', 839, 1, 'G', 'SNP', 2, -1, -1, 3, 66, -1, -1] == ms.loc['reference:839:1:G'].values.tolist()
    assert ['reference', 866, 9, 'G', 'INDEL', 1, -1, -1, 3, 33, -1, -1] == ms.loc['reference:866:9:G'].values.tolist()
    assert ['reference', 1048, 1, 'G', 'SNP', 1, -1, -1, 3, 33, -1, -1] == ms.loc['reference:1048:1:G'].values.tolist()
    assert ['reference', 3897, 5, 'G', 'INDEL', 2, -1, -1, 3, 66, -1, -1] == ms.loc[
        'reference:3897:5:G'].values.tolist()

    # Mutations include unknown spdi
    ms = gds.mutations_summary('genome', id_type='spdi', ignore_annotations=True, include_unknown_features=True,
                               include_unknown_samples=False)
    assert 112 + 440 == len(ms)
    assert 'Mutation' == ms.index.name
    assert ['Sequence', 'Position', 'Deletion', 'Insertion', 'Type',
            'Count', 'Unknown Count', 'Present and Unknown Count', 'Total',
            'Percent', 'Unknown Percent', 'Present and Unknown Percent'] == list(ms.columns)

    ## Convert percent to int to make it easier to compare in assert statements
    ms['Percent'] = ms['Percent'].astype(int)
    ms['Unknown Count'] = ms['Unknown Count'].fillna(-1).astype(int)
    ms['Present and Unknown Count'] = ms['Present and Unknown Count'].fillna(-1).astype(int)
    ms['Unknown Percent'] = ms['Unknown Percent'].fillna(-1).astype(int)
    ms['Present and Unknown Percent'] = ms['Present and Unknown Percent'].fillna(-1).astype(int)

    assert ['reference', 839, 1, 'G', 'SNP', 2, -1, -1, 3, 66, -1, -1] == ms.loc['reference:839:1:G'].values.tolist()
    assert ['reference', 866, 9, 'G', 'INDEL', 1, -1, -1, 3, 33, -1, -1] == ms.loc['reference:866:9:G'].values.tolist()
    assert ['reference', 1048, 1, 'G', 'SNP', 1, -1, -1, 3, 33, -1, -1] == ms.loc['reference:1048:1:G'].values.tolist()
    assert ['reference', 3897, 5, 'G', 'INDEL', 2, -1, -1, 3, 66, -1, -1] == ms.loc[
        'reference:3897:5:G'].values.tolist()
    assert ['reference', 89, 1, '?', 'UNKNOWN_MISSING', 3, -1, -1, 3, 100, -1, -1] == ms.loc[
        'reference:89:1:?'].values.tolist()
    assert ['reference', 5100, 1, '?', 'UNKNOWN_MISSING', 2, -1, -1, 3, 66, -1, -1] == ms.loc[
        'reference:5100:1:?'].values.tolist()
    assert ['reference', 649, 1, '?', 'UNKNOWN_MISSING', 1, -1, -1, 3, 33, -1, -1] == ms.loc[
        'reference:649:1:?'].values.tolist()

    # Mutations spdi ref
    ms = gds.mutations_summary('genome', id_type='spdi_ref', ignore_annotations=True,
                               include_unknown_samples=False)
    assert 112 == len(ms)
    assert 'Mutation' == ms.index.name
    assert ['Sequence', 'Position', 'Deletion', 'Insertion', 'Type',
            'Count', 'Unknown Count', 'Present and Unknown Count', 'Total',
            'Percent', 'Unknown Percent', 'Present and Unknown Percent'] == list(ms.columns)

    ## Convert percent to int to make it easier to compare in assert statements
    ms['Percent'] = ms['Percent'].astype(int)
    ms['Unknown Count'] = ms['Unknown Count'].fillna(-1).astype(int)
    ms['Present and Unknown Count'] = ms['Present and Unknown Count'].fillna(-1).astype(int)
    ms['Unknown Percent'] = ms['Unknown Percent'].fillna(-1).astype(int)
    ms['Present and Unknown Percent'] = ms['Present and Unknown Percent'].fillna(-1).astype(int)

    assert ['reference', 839, 'C', 'G', 'SNP', 2, -1, -1, 3, 66, -1, -1] == ms.loc['reference:839:C:G'].values.tolist()
    assert ['reference', 866, 'GCCAGATCC', 'G', 'INDEL', 1, -1, -1, 3, 33, -1, -1] == ms.loc[
        'reference:866:GCCAGATCC:G'].values.tolist()
    assert ['reference', 1048, 'C', 'G', 'SNP', 1, -1, -1, 3, 33, -1, -1] == ms.loc[
        'reference:1048:C:G'].values.tolist()
    assert ['reference', 3897, 'GCGCA', 'G', 'INDEL', 2, -1, -1, 3, 66, -1, -1] == ms.loc[
        'reference:3897:GCGCA:G'].values.tolist()

    # Mutations spdi ref include unknowns
    ms = gds.mutations_summary('genome', id_type='spdi_ref', ignore_annotations=True, include_unknown_features=True,
                               include_unknown_samples=False)
    assert 112 + 440 == len(ms)
    assert 'Mutation' == ms.index.name
    assert ['Sequence', 'Position', 'Deletion', 'Insertion', 'Type',
            'Count', 'Unknown Count', 'Present and Unknown Count', 'Total',
            'Percent', 'Unknown Percent', 'Present and Unknown Percent'] == list(ms.columns)

    ## Convert percent to int to make it easier to compare in assert statements
    ms['Percent'] = ms['Percent'].astype(int)
    ms['Unknown Count'] = ms['Unknown Count'].fillna(-1).astype(int)
    ms['Present and Unknown Count'] = ms['Present and Unknown Count'].fillna(-1).astype(int)
    ms['Unknown Percent'] = ms['Unknown Percent'].fillna(-1).astype(int)
    ms['Present and Unknown Percent'] = ms['Present and Unknown Percent'].fillna(-1).astype(int)

    assert ['reference', 839, 'C', 'G', 'SNP', 2, -1, -1, 3, 66, -1, -1] == ms.loc['reference:839:C:G'].values.tolist()
    assert ['reference', 866, 'GCCAGATCC', 'G', 'INDEL', 1, -1, -1, 3, 33, -1, -1] == ms.loc[
        'reference:866:GCCAGATCC:G'].values.tolist()
    assert ['reference', 1048, 'C', 'G', 'SNP', 1, -1, -1, 3, 33, -1, -1] == ms.loc[
        'reference:1048:C:G'].values.tolist()
    assert ['reference', 3897, 'GCGCA', 'G', 'INDEL', 2, -1, -1, 3, 66, -1, -1] == ms.loc[
        'reference:3897:GCGCA:G'].values.tolist()
    assert ['reference', 89, 'A', '?', 'UNKNOWN_MISSING', 3, -1, -1, 3, 100, -1, -1] == ms.loc[
        'reference:89:A:?'].values.tolist()
    assert ['reference', 5100, 'T', '?', 'UNKNOWN_MISSING', 2, -1, -1, 3, 66, -1, -1] == ms.loc[
        'reference:5100:T:?'].values.tolist()
    assert ['reference', 649, 'T', '?', 'UNKNOWN_MISSING', 1, -1, -1, 3, 33, -1, -1] == ms.loc[
        'reference:649:T:?'].values.tolist()

    # Mutations include annotations (which should all be empty)
    ms = gds.mutations_summary('genome', id_type='spdi', ignore_annotations=False,
                               include_unknown_samples=False)
    assert 112 == len(ms)
    assert 'Mutation' == ms.index.name
    assert ['Sequence', 'Position', 'Deletion', 'Insertion', 'Type',
            'Count', 'Unknown Count', 'Present and Unknown Count', 'Total',
            'Percent', 'Unknown Percent', 'Present and Unknown Percent',
            'Annotation', 'Annotation_Impact',
            'Gene_Name', 'Gene_ID', 'Feature_Type', 'Transcript_BioType',
            'HGVS.c', 'HGVS.p', 'ID_HGVS.c', 'ID_HGVS.p', 'ID_HGVS_GN.c', 'ID_HGVS_GN.p'] == list(ms.columns)

    ## Convert percent to int to make it easier to compare in assert statements
    ms['Percent'] = ms['Percent'].astype(int)
    ms['Unknown Count'] = ms['Unknown Count'].fillna(-1).astype(int)
    ms['Present and Unknown Count'] = ms['Present and Unknown Count'].fillna(-1).astype(int)
    ms['Unknown Percent'] = ms['Unknown Percent'].fillna(-1).astype(int)
    ms['Present and Unknown Percent'] = ms['Present and Unknown Percent'].fillna(-1).astype(int)

    assert ['reference', 839, 1, 'G', 'SNP', 2, -1, -1, 3, 66, -1, -1] + ['NA'] * 12 == ms.loc[
        'reference:839:1:G'].fillna(
        'NA').values.tolist()

    # Test case of directly calling features_summary
    ms = gds.features_summary(kind='mutations', scope='genome', id_type='spdi', ignore_annotations=True,
                              include_unknown_samples=False)
    assert 112 == len(ms)
    assert 'Mutation' == ms.index.name
    assert ['Sequence', 'Position', 'Deletion', 'Insertion', 'Type',
            'Count', 'Unknown Count', 'Present and Unknown Count', 'Total',
            'Percent', 'Unknown Percent', 'Present and Unknown Percent'] == list(ms.columns)

    ## Convert percent to int to make it easier to compare in assert statements
    ms['Percent'] = ms['Percent'].astype(int)
    ms['Unknown Count'] = ms['Unknown Count'].fillna(-1).astype(int)
    ms['Present and Unknown Count'] = ms['Present and Unknown Count'].fillna(-1).astype(int)
    ms['Unknown Percent'] = ms['Unknown Percent'].fillna(-1).astype(int)
    ms['Present and Unknown Percent'] = ms['Present and Unknown Percent'].fillna(-1).astype(int)

    assert ['reference', 839, 1, 'G', 'SNP', 2, -1, -1, 3, 66, -1, -1] == ms.loc['reference:839:1:G'].values.tolist()
    assert ['reference', 866, 9, 'G', 'INDEL', 1, -1, -1, 3, 33, -1, -1] == ms.loc['reference:866:9:G'].values.tolist()
    assert ['reference', 1048, 1, 'G', 'SNP', 1, -1, -1, 3, 33, -1, -1] == ms.loc['reference:1048:1:G'].values.tolist()
    assert ['reference', 3897, 5, 'G', 'INDEL', 2, -1, -1, 3, 66, -1, -1] == ms.loc[
        'reference:3897:5:G'].values.tolist()

    # Test case of only including unknowns
    ms = gds.mutations_summary('genome', id_type='spdi', include_present_features=False, include_unknown_features=True,
                               include_unknown_samples=False)
    assert 440 == len(ms)
    assert 'reference:649:1:?' in set(ms.index.tolist())
    assert 'reference:839:1:G' not in (ms.index.tolist())

    # Test case of no present or unknowns
    ms = gds.mutations_summary('genome', id_type='spdi', include_present_features=False, include_unknown_features=False,
                               include_unknown_samples=False)
    assert 0 == len(ms)


def test_summaries_mlst_data(loaded_database_genomic_data_store: GenomicsDataIndex):
    gds = loaded_database_genomic_data_store

    # MLST summaries for lmonocytogenes
    summary_df = gds.features_summary(kind='mlst', scope='lmonocytogenes', include_unknown_samples=True)
    summary_df['Percent'] = summary_df['Percent'].astype(int)  # Convert to int for easier comparison
    summary_df['Unknown Count'] = summary_df['Unknown Count'].fillna(-1).astype(int)
    summary_df['Present and Unknown Count'] = summary_df['Present and Unknown Count'].fillna(-1).astype(int)
    summary_df['Unknown Percent'] = summary_df['Unknown Percent'].fillna(-1).astype(int)
    summary_df['Present and Unknown Percent'] = summary_df['Present and Unknown Percent'].fillna(-1).astype(int)
    assert 10 == len(summary_df)
    assert 'MLST Feature' == summary_df.index.name
    assert ['Scheme', 'Locus', 'Allele', 'Count', 'Unknown Count', 'Present and Unknown Count',
            'Total', 'Percent', 'Unknown Percent', 'Present and Unknown Percent'] == list(summary_df.columns)
    assert ['lmonocytogenes', 'abcZ', '1', 5, 0, 5, 5, 100, 0, 100] == summary_df.loc[
        'mlst:lmonocytogenes:abcZ:1'].tolist()
    assert ['lmonocytogenes', 'bglA', '51', 3, 0, 3, 5, 60, 0, 60] == summary_df.loc[
        'mlst:lmonocytogenes:bglA:51'].tolist()
    assert ['lmonocytogenes', 'lhkA', '4', 1, 0, 1, 5, 20, 0, 20] == summary_df.loc[
        'mlst:lmonocytogenes:lhkA:4'].tolist()
    assert ['lmonocytogenes', 'lhkA', '5', 4, 0, 4, 5, 80, 0, 80] == summary_df.loc[
        'mlst:lmonocytogenes:lhkA:5'].tolist()
    assert ['lmonocytogenes', 'ldh', '5', 4, 1, 5, 5, 80, 20, 100] == summary_df.loc[
        'mlst:lmonocytogenes:ldh:5'].tolist()

    # MLST summaries for lmonocytogenes include unknown
    summary_df = gds.features_summary(kind='mlst', scope='lmonocytogenes', include_unknown_features=True,
                                      include_unknown_samples=True)
    summary_df['Percent'] = summary_df['Percent'].astype(int)  # Convert to int for easier comparison
    summary_df['Unknown Count'] = summary_df['Unknown Count'].fillna(-1).astype(int)
    summary_df['Present and Unknown Count'] = summary_df['Present and Unknown Count'].fillna(-1).astype(int)
    summary_df['Unknown Percent'] = summary_df['Unknown Percent'].fillna(-1).astype(int)
    summary_df['Present and Unknown Percent'] = summary_df['Present and Unknown Percent'].fillna(-1).astype(int)
    assert 11 == len(summary_df)
    assert 'MLST Feature' == summary_df.index.name
    assert ['Scheme', 'Locus', 'Allele', 'Count', 'Unknown Count', 'Present and Unknown Count',
            'Total', 'Percent', 'Unknown Percent', 'Present and Unknown Percent'] == list(summary_df.columns)
    assert ['lmonocytogenes', 'abcZ', '1', 5, 0, 5, 5, 100, 0, 100] == summary_df.loc[
        'mlst:lmonocytogenes:abcZ:1'].tolist()
    assert ['lmonocytogenes', 'bglA', '51', 3, 0, 3, 5, 60, 0, 60] == summary_df.loc[
        'mlst:lmonocytogenes:bglA:51'].tolist()
    assert ['lmonocytogenes', 'lhkA', '4', 1, 0, 1, 5, 20, 0, 20] == summary_df.loc[
        'mlst:lmonocytogenes:lhkA:4'].tolist()
    assert ['lmonocytogenes', 'lhkA', '5', 4, 0, 4, 5, 80, 0, 80] == summary_df.loc[
        'mlst:lmonocytogenes:lhkA:5'].tolist()
    assert ['lmonocytogenes', 'ldh', '5', 4, 1, 5, 5, 80, 20, 100] == summary_df.loc[
        'mlst:lmonocytogenes:ldh:5'].tolist()
    assert ['lmonocytogenes', 'ldh', '?', 1, -1, -1, 5, 20, -1, -1] == summary_df.loc[
        'mlst:lmonocytogenes:ldh:?'].tolist()

    # MLST summaries for lmonocytogenes with specific locus id
    summary_df = gds.features_summary(kind='mlst', scope='lmonocytogenes', locus='bglA', include_unknown_samples=True)
    summary_df['Percent'] = summary_df['Percent'].astype(int)  # Convert to int for easier comparison
    summary_df['Unknown Count'] = summary_df['Unknown Count'].fillna(-1).astype(int)
    summary_df['Present and Unknown Count'] = summary_df['Present and Unknown Count'].fillna(-1).astype(int)
    summary_df['Unknown Percent'] = summary_df['Unknown Percent'].fillna(-1).astype(int)
    summary_df['Present and Unknown Percent'] = summary_df['Present and Unknown Percent'].fillna(-1).astype(int)
    assert 2 == len(summary_df)
    assert 'MLST Feature' == summary_df.index.name
    assert ['Scheme', 'Locus', 'Allele', 'Count', 'Unknown Count', 'Present and Unknown Count',
            'Total', 'Percent', 'Unknown Percent', 'Present and Unknown Percent'] == list(summary_df.columns)
    assert ['lmonocytogenes', 'bglA', '51', 3, 0, 3, 5, 60, 0, 60] == summary_df.loc[
        'mlst:lmonocytogenes:bglA:51'].tolist()
    assert ['lmonocytogenes', 'bglA', '52', 2, 0, 2, 5, 40, 0, 40] == summary_df.loc[
        'mlst:lmonocytogenes:bglA:52'].tolist()

    # MLST summaries for lmonocytogenes include unknown and not present
    summary_df = gds.features_summary(kind='mlst', scope='lmonocytogenes', include_present_features=False,
                                      include_unknown_features=True, include_unknown_samples=True)
    summary_df['Percent'] = summary_df['Percent'].astype(int)  # Convert to int for easier comparison
    summary_df['Unknown Count'] = summary_df['Unknown Count'].fillna(-1).astype(int)
    summary_df['Present and Unknown Count'] = summary_df['Present and Unknown Count'].fillna(-1).astype(int)
    summary_df['Unknown Percent'] = summary_df['Unknown Percent'].fillna(-1).astype(int)
    summary_df['Present and Unknown Percent'] = summary_df['Present and Unknown Percent'].fillna(-1).astype(int)
    assert 1 == len(summary_df)
    assert 'MLST Feature' == summary_df.index.name
    assert ['Scheme', 'Locus', 'Allele', 'Count', 'Unknown Count', 'Present and Unknown Count',
            'Total', 'Percent', 'Unknown Percent', 'Present and Unknown Percent'] == list(summary_df.columns)
    assert ['lmonocytogenes', 'ldh', '?', 1, -1, -1, 5, 20, -1, -1] == summary_df.loc[
        'mlst:lmonocytogenes:ldh:?'].tolist()

    # MLST summaries for lmonocytogenes not include present or unknown
    summary_df = gds.features_summary(kind='mlst', scope='lmonocytogenes', include_present_features=False,
                                      include_unknown_features=False, include_unknown_samples=True)
    summary_df['Percent'] = summary_df['Percent'].astype(int)  # Convert to int for easier comparison
    summary_df['Unknown Count'] = summary_df['Unknown Count'].fillna(-1).astype(int)
    summary_df['Present and Unknown Count'] = summary_df['Present and Unknown Count'].fillna(-1).astype(int)
    summary_df['Unknown Percent'] = summary_df['Unknown Percent'].fillna(-1).astype(int)
    summary_df['Present and Unknown Percent'] = summary_df['Present and Unknown Percent'].fillna(-1).astype(int)
    assert 0 == len(summary_df)
    assert 'MLST Feature' == summary_df.index.name
    assert ['Scheme', 'Locus', 'Allele', 'Count', 'Unknown Count', 'Present and Unknown Count',
            'Total', 'Percent', 'Unknown Percent', 'Present and Unknown Percent'] == list(summary_df.columns)

    # Summaries using 'mlst_summery()'
    summary_df = gds.mlst_summary(scheme_name='lmonocytogenes', include_unknown_samples=True)
    assert 10 == len(summary_df)


def test_summaries_variant_annotations(
        loaded_database_genomic_data_store_annotations_include_unknown: GenomicsDataIndex):
    gds = loaded_database_genomic_data_store_annotations_include_unknown

    # Samples
    assert 3 == gds.count_samples()
    assert {'SH10-014', 'SH14-014', 'SH14-001'} == set(gds.sample_names())

    # References
    assert 1 == gds.count_references()
    assert ['NC_011083'] == gds.reference_names()

    # Mutations
    assert 177 == gds.count_mutations('NC_011083')

    # spdi
    ms = gds.mutations_summary('NC_011083', id_type='spdi', ignore_annotations=False, include_unknown_samples=True)
    assert 177 == len(ms)
    assert 'Mutation' == ms.index.name
    assert ['Sequence', 'Position', 'Deletion', 'Insertion', 'Type',
            'Count', 'Unknown Count', 'Present and Unknown Count', 'Total',
            'Percent', 'Unknown Percent', 'Present and Unknown Percent',
            'Annotation', 'Annotation_Impact',
            'Gene_Name', 'Gene_ID', 'Feature_Type', 'Transcript_BioType',
            'HGVS.c', 'HGVS.p', 'ID_HGVS.c', 'ID_HGVS.p', 'ID_HGVS_GN.c', 'ID_HGVS_GN.p'] == list(ms.columns)

    ## Convert percent to int to make it easier to compare in assert statements
    ms['Percent'] = ms['Percent'].astype(int)
    ms['Unknown Count'] = ms['Unknown Count'].fillna(-1).astype(int)
    ms['Present and Unknown Count'] = ms['Present and Unknown Count'].fillna(-1).astype(int)
    ms['Unknown Percent'] = ms['Unknown Percent'].fillna(-1).astype(int)
    ms['Present and Unknown Percent'] = ms['Present and Unknown Percent'].fillna(-1).astype(int)

    ## missense variant
    assert ['NC_011083', 140658, 1, 'A', 'SNP', 3, 0, 3, 3, 100, 0, 100,
            'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu',
            'hgvs:NC_011083:SEHA_RS01180:c.497C>A', 'hgvs:NC_011083:SEHA_RS01180:p.Ala166Glu',
            'hgvs_gn:NC_011083:murF:c.497C>A', 'hgvs_gn:NC_011083:murF:p.Ala166Glu'] == list(
        ms.loc['NC_011083:140658:1:A'])

    ## inframe deletion
    assert ['NC_011083', 4465400, len('GGCCGAA'), 'G', 'INDEL', 3, 0, 3, 3, 100, 0, 100,
            'conservative_inframe_deletion', 'MODERATE', 'tyrB', 'SEHA_RS22180', 'transcript', 'protein_coding',
            'c.157_162delGAAGCC', 'p.Glu53_Ala54del',
            'hgvs:NC_011083:SEHA_RS22180:c.157_162delGAAGCC', 'hgvs:NC_011083:SEHA_RS22180:p.Glu53_Ala54del',
            'hgvs_gn:NC_011083:tyrB:c.157_162delGAAGCC', 'hgvs_gn:NC_011083:tyrB:p.Glu53_Ala54del'] == list(
        ms.loc['NC_011083:4465400:7:G'])

    ## Intergenic variant (with some NA values in fields)
    assert ['NC_011083', 4555461, len('T'), 'TC', 'INDEL', 1, 2, 3, 3, 33, 66, 100,
            'intergenic_region', 'MODIFIER', 'SEHA_RS22510-SEHA_RS26685', 'SEHA_RS22510-SEHA_RS26685',
            'intergenic_region', 'NA',
            'n.4555461_4555462insC', 'NA',
            'hgvs:NC_011083:n.4555461_4555462insC', 'NA',
            'hgvs_gn:NC_011083:n.4555461_4555462insC', 'NA'] == list(ms.loc['NC_011083:4555461:1:TC'].fillna('NA'))

    ## MNP variant (2/3, 1/3)
    assert ['NC_011083', 3535698, len('GCC'), 'CAT', 'MNP', 2, 1, 3, 3, 66, 33, 100,
            'missense_variant', 'MODERATE', 'oadA', 'SEHA_RS17780',
            'transcript', 'protein_coding',
            'c.544_546delGGCinsATG', 'p.Gly182Met',
            'hgvs:NC_011083:SEHA_RS17780:c.544_546delGGCinsATG', 'hgvs:NC_011083:SEHA_RS17780:p.Gly182Met',
            'hgvs_gn:NC_011083:oadA:c.544_546delGGCinsATG', 'hgvs_gn:NC_011083:oadA:p.Gly182Met'] == list(
        ms.loc['NC_011083:3535698:3:CAT'])

    ## Long MNP variant (2/3, 1/3)
    assert ['NC_011083', 3535143, len('AATGCCTGCC'), 'TATCCCGGCG', 'MNP', 2, 1, 3, 3, 66, 33, 100,
            'synonymous_variant', 'LOW', 'oadA', 'SEHA_RS17780', 'transcript', 'protein_coding',
            'c.1092_1101delGGCAGGCATTinsCGCCGGGATA', 'p.368',
            'hgvs:NC_011083:SEHA_RS17780:c.1092_1101delGGCAGGCATTinsCGCCGGGATA',
            'hgvs:NC_011083:SEHA_RS17780:p.368',
            'hgvs_gn:NC_011083:oadA:c.1092_1101delGGCAGGCATTinsCGCCGGGATA',
            'hgvs_gn:NC_011083:oadA:p.368'] == list(
        ms.loc[f'NC_011083:3535143:{len("AATGCCTGCC")}:TATCCCGGCG'])

    ## synonymous variant, no unknown (1/3, 0/3)
    assert ['NC_011083', 508378, len('C'), 'T', 'SNP', 1, 0, 1, 3, 33, 0, 33,
            'synonymous_variant', 'LOW', 'tgt', 'SEHA_RS02965', 'transcript', 'protein_coding',
            'c.423C>T', 'p.Ile141Ile',
            'hgvs:NC_011083:SEHA_RS02965:c.423C>T', 'hgvs:NC_011083:SEHA_RS02965:p.Ile141Ile',
            'hgvs_gn:NC_011083:tgt:c.423C>T', 'hgvs_gn:NC_011083:tgt:p.Ile141Ile'] == list(
        ms.loc['NC_011083:508378:1:T'])

    # deletion where there is an overlap with present and unknown, (present 3/3, unknown 2/3,
    # overlap 1676762 and 1676763)
    assert ['NC_011083', 1676762, len('CA'), 'C', 'INDEL', 1, 2, 3, 3, 33, 66, 100,
            'intergenic_region', 'MODIFIER', 'SEHA_RS26130-SEHA_RS08880', 'SEHA_RS26130-SEHA_RS08880',
            'intergenic_region', '<NA>',
            'n.1676763delA', '<NA>',
            'hgvs:NC_011083:n.1676763delA', '<NA>',
            'hgvs_gn:NC_011083:n.1676763delA', '<NA>'] == list(
        ms.loc['NC_011083:1676762:2:C'].fillna('<NA>'))

    # spdi_ref
    ms = gds.mutations_summary('NC_011083', id_type='spdi_ref', include_unknown_samples=True)
    assert 177 == len(ms)
    assert 'Mutation' == ms.index.name
    assert ['Sequence', 'Position', 'Deletion', 'Insertion', 'Type',
            'Count', 'Unknown Count', 'Present and Unknown Count', 'Total',
            'Percent', 'Unknown Percent', 'Present and Unknown Percent',
            'Annotation', 'Annotation_Impact',
            'Gene_Name', 'Gene_ID', 'Feature_Type', 'Transcript_BioType',
            'HGVS.c', 'HGVS.p', 'ID_HGVS.c', 'ID_HGVS.p', 'ID_HGVS_GN.c', 'ID_HGVS_GN.p'] == list(ms.columns)

    ## Convert percent to int to make it easier to compare in assert statements
    ms['Percent'] = ms['Percent'].astype(int)
    ms['Unknown Count'] = ms['Unknown Count'].fillna(-1).astype(int)
    ms['Present and Unknown Count'] = ms['Present and Unknown Count'].fillna(-1).astype(int)
    ms['Unknown Percent'] = ms['Unknown Percent'].fillna(-1).astype(int)
    ms['Present and Unknown Percent'] = ms['Present and Unknown Percent'].fillna(-1).astype(int)

    ## missense variant
    assert ['NC_011083', 140658, 'C', 'A', 'SNP', 3, 0, 3, 3, 100, 0, 100,
            'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu',
            'hgvs:NC_011083:SEHA_RS01180:c.497C>A', 'hgvs:NC_011083:SEHA_RS01180:p.Ala166Glu',
            'hgvs_gn:NC_011083:murF:c.497C>A', 'hgvs_gn:NC_011083:murF:p.Ala166Glu'] == list(
        ms.loc['NC_011083:140658:C:A'])

    ## inframe deletion
    assert ['NC_011083', 4465400, 'GGCCGAA', 'G', 'INDEL', 3, 0, 3, 3, 100, 0, 100,
            'conservative_inframe_deletion', 'MODERATE', 'tyrB', 'SEHA_RS22180', 'transcript', 'protein_coding',
            'c.157_162delGAAGCC', 'p.Glu53_Ala54del',
            'hgvs:NC_011083:SEHA_RS22180:c.157_162delGAAGCC', 'hgvs:NC_011083:SEHA_RS22180:p.Glu53_Ala54del',
            'hgvs_gn:NC_011083:tyrB:c.157_162delGAAGCC', 'hgvs_gn:NC_011083:tyrB:p.Glu53_Ala54del'] == list(
        ms.loc['NC_011083:4465400:GGCCGAA:G'])

    ## Intergenic variant (with some NA values in fields)
    assert ['NC_011083', 4555461, 'T', 'TC', 'INDEL', 1, 2, 3, 3, 33, 66, 100,
            'intergenic_region', 'MODIFIER', 'SEHA_RS22510-SEHA_RS26685', 'SEHA_RS22510-SEHA_RS26685',
            'intergenic_region', 'NA',
            'n.4555461_4555462insC', 'NA',
            'hgvs:NC_011083:n.4555461_4555462insC', 'NA',
            'hgvs_gn:NC_011083:n.4555461_4555462insC', 'NA'] == list(ms.loc['NC_011083:4555461:T:TC'].fillna('NA'))

    ## MNP variant (2/3, 1/3)
    assert ['NC_011083', 3535698, 'GCC', 'CAT', 'MNP', 2, 1, 3, 3, 66, 33, 100,
            'missense_variant', 'MODERATE', 'oadA', 'SEHA_RS17780',
            'transcript', 'protein_coding',
            'c.544_546delGGCinsATG', 'p.Gly182Met',
            'hgvs:NC_011083:SEHA_RS17780:c.544_546delGGCinsATG', 'hgvs:NC_011083:SEHA_RS17780:p.Gly182Met',
            'hgvs_gn:NC_011083:oadA:c.544_546delGGCinsATG', 'hgvs_gn:NC_011083:oadA:p.Gly182Met'] == list(
        ms.loc['NC_011083:3535698:GCC:CAT'])

    ## Long MNP variant (2/3, 1/3)
    assert ['NC_011083', 3535143, 'AATGCCTGCC', 'TATCCCGGCG', 'MNP', 2, 1, 3, 3, 66, 33, 100,
            'synonymous_variant', 'LOW', 'oadA', 'SEHA_RS17780', 'transcript', 'protein_coding',
            'c.1092_1101delGGCAGGCATTinsCGCCGGGATA', 'p.368',
            'hgvs:NC_011083:SEHA_RS17780:c.1092_1101delGGCAGGCATTinsCGCCGGGATA',
            'hgvs:NC_011083:SEHA_RS17780:p.368',
            'hgvs_gn:NC_011083:oadA:c.1092_1101delGGCAGGCATTinsCGCCGGGATA',
            'hgvs_gn:NC_011083:oadA:p.368'] == list(
        ms.loc['NC_011083:3535143:AATGCCTGCC:TATCCCGGCG'])

    ## synonymous variant, no unknown (1/3, 0/3)
    assert ['NC_011083', 508378, 'C', 'T', 'SNP', 1, 0, 1, 3, 33, 0, 33,
            'synonymous_variant', 'LOW', 'tgt', 'SEHA_RS02965', 'transcript', 'protein_coding',
            'c.423C>T', 'p.Ile141Ile',
            'hgvs:NC_011083:SEHA_RS02965:c.423C>T', 'hgvs:NC_011083:SEHA_RS02965:p.Ile141Ile',
            'hgvs_gn:NC_011083:tgt:c.423C>T', 'hgvs_gn:NC_011083:tgt:p.Ile141Ile'] == list(
        ms.loc['NC_011083:508378:C:T'])

    # deletion where there is an overlap with present and unknown, (present 3/3, unknown 2/3,
    # overlap 1676762 and 1676763)
    assert ['NC_011083', 1676762, 'CA', 'C', 'INDEL', 1, 2, 3, 3, 33, 66, 100,
            'intergenic_region', 'MODIFIER', 'SEHA_RS26130-SEHA_RS08880', 'SEHA_RS26130-SEHA_RS08880',
            'intergenic_region', '<NA>',
            'n.1676763delA', '<NA>',
            'hgvs:NC_011083:n.1676763delA', '<NA>',
            'hgvs_gn:NC_011083:n.1676763delA', '<NA>'] == list(
        ms.loc['NC_011083:1676762:CA:C'].fillna('<NA>'))


def test_connect_to_project_from_dir():
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        project_dir = tmp_file / 'project'
        Project.initialize_project(project_dir)

        ds = GenomicsDataIndex.connect(project_dir=project_dir)
        assert ds is not None
        assert ds.connection.reference_service is not None
        assert ds.connection.filesystem_storage.variation_dir.parent == project_dir / '.gdi-data'


def test_connect_to_project_from_project():
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        project_dir = tmp_file / 'project'
        project = Project.initialize_project(project_dir)

        ds = GenomicsDataIndex.connect(project=project)
        assert ds is not None
        assert ds.connection.reference_service is not None
        assert ds.connection.filesystem_storage.variation_dir.parent == project_dir / '.gdi-data'


def test_connect_and_tree_after_moving_project(loaded_data_store_from_project_dir):
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        project_dir = tmp_file / 'project'
        Project.initialize_project(project_dir)
        ds = GenomicsDataIndex.connect(project_dir=project_dir)
        database_connection = ds.connection

        # Load Nucleotide variation
        database_connection.reference_service.add_reference_genome(reference_file)
        snippy_tmp_dir = Path(tempfile.mkdtemp())
        data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                                      SerialSampleFilesProcessor(snippy_tmp_dir))
        database_connection.variation_service.insert(data_package, feature_scope_name='genome')

        # I should be able to build tree initially
        tq1 = ds.samples_query().build_tree(kind='mutation', scope='genome')
        assert tq1.tree is not None

        # Move project
        project_dir_2 = tmp_file / 'project2'
        shutil.move(project_dir, project_dir_2)

        ds2 = GenomicsDataIndex.connect(project_dir=project_dir_2)

        # I should still be able to build tree even after project has been moved
        tq2 = ds2.samples_query().build_tree(kind='mutation', scope='genome')
        assert tq2.tree is not None



import pytest

from genomics_data_index.storage.model.db import SampleMLSTAlleles, MLSTAllelesSamples, Sample, MLSTScheme
from genomics_data_index.storage.service import EntityExistsError
from genomics_data_index.storage.service.MLSTService import MLSTService


def test_insert_mlst_results(database, mlst_data_package_single_scheme, sample_service, filesystem_storage):
    num_loci = 7

    mlst_service = MLSTService(database_connection=database,
                               sample_service=sample_service,
                               mlst_dir=filesystem_storage.mlst_dir)

    session = database.get_session()

    mlst_service.insert(feature_scope_name='lmonocytogenes', data_package=mlst_data_package_single_scheme)

    samples = session.query(Sample).all()

    assert 2 == len(samples)
    assert {'CFSAN002349', 'CFSAN023463'} == {s.name for s in samples}

    sample_mlst_alleles = session.query(SampleMLSTAlleles).all()
    assert 2 == len(sample_mlst_alleles)

    mlst_alleles_samples = session.query(MLSTAllelesSamples).all()
    mlst_alleles_samples_id_allele = {allele.sla: allele for allele in mlst_alleles_samples}

    assert num_loci == len(mlst_alleles_samples)
    assert {'lmonocytogenes:abcZ:1', 'lmonocytogenes:bglA:51', 'lmonocytogenes:cat:11',
            'lmonocytogenes:dapE:13', 'lmonocytogenes:dat:2', 'lmonocytogenes:ldh:5',
            'lmonocytogenes:lhkA:5'} == set(mlst_alleles_samples_id_allele.keys())

    assert 2 == len(mlst_alleles_samples_id_allele['lmonocytogenes:abcZ:1'].sample_ids)


def test_insert_mlst_results_multiple_schemes(database, mlst_data_package_basic, sample_service, filesystem_storage):
    num_loci = 7
    num_schemes = 3

    mlst_service = MLSTService(database_connection=database,
                               sample_service=sample_service,
                               mlst_dir=filesystem_storage.mlst_dir)

    session = database.get_session()

    mlst_service.insert(data_package=mlst_data_package_basic)

    samples = session.query(Sample).all()

    assert 6 == len(samples)
    assert {'CFSAN002349', 'CFSAN023463', '2014C-3598', '2014C-3599',
            '2014D-0067', '2014D-0068'} == {s.name for s in samples}

    sample_mlst_alleles = session.query(SampleMLSTAlleles).all()
    assert 6 == len(sample_mlst_alleles)

    mlst_alleles_all = session.query(MLSTAllelesSamples).all()
    mlst_alleles_all_id = {allele.sla: allele for allele in mlst_alleles_all}
    assert num_loci * num_schemes + 2 == len(mlst_alleles_all)

    assert {'lmonocytogenes', 'ecoli', 'campylobacter'} == {allele.scheme for allele in mlst_alleles_all}

    assert 2 == len(mlst_alleles_all_id['lmonocytogenes:abcZ:1'].sample_ids)
    assert 2 == len(mlst_alleles_all_id['ecoli:fumC:23'].sample_ids)

    mlst_alleles_lmono = session.query(MLSTAllelesSamples).filter(MLSTAllelesSamples.scheme == 'lmonocytogenes').all()
    mlst_alleles_id_lmono = {allele.sla for allele in mlst_alleles_lmono}

    assert {'lmonocytogenes:abcZ:1', 'lmonocytogenes:bglA:51', 'lmonocytogenes:cat:11',
            'lmonocytogenes:dapE:13', 'lmonocytogenes:dat:2', 'lmonocytogenes:ldh:5',
            'lmonocytogenes:lhkA:4', 'lmonocytogenes:lhkA:5'} == mlst_alleles_id_lmono


def test_insert_mlst_results_multiple_schemes_override_scheme(database, mlst_data_package_basic,
                                                              sample_service, filesystem_storage):
    num_loci = 7
    num_alt_schemes = 3

    mlst_service = MLSTService(database_connection=database,
                               sample_service=sample_service,
                               mlst_dir=filesystem_storage.mlst_dir)

    session = database.get_session()

    mlst_service.insert(feature_scope_name='lmonocytogenes', data_package=mlst_data_package_basic)

    sample_mlst_alleles = session.query(SampleMLSTAlleles).all()
    assert 6 == len(sample_mlst_alleles)

    mlst_alleles_samples = session.query(MLSTAllelesSamples).all()
    mlst_alleles_samples_id_allele = {allele.sla: allele for allele in mlst_alleles_samples}

    assert num_loci * num_alt_schemes + 2 == len(mlst_alleles_samples)
    assert 2 == len(mlst_alleles_samples_id_allele['lmonocytogenes:abcZ:1'].sample_ids)
    assert 2 == len(mlst_alleles_samples_id_allele['lmonocytogenes:fumC:23'].sample_ids)


def test_double_insert_mlst(database, mlst_data_package_single_scheme, sample_service, filesystem_storage):
    mlst_service = MLSTService(database_connection=database,
                               sample_service=sample_service,
                               mlst_dir=filesystem_storage.mlst_dir)

    session = database.get_session()

    mlst_service.insert(feature_scope_name='lmonocytogenes', data_package=mlst_data_package_single_scheme)

    samples = session.query(Sample).all()

    assert 2 == len(samples)
    assert {'CFSAN002349', 'CFSAN023463'} == {s.name for s in samples}

    sample_mlst_alleles = session.query(SampleMLSTAlleles).all()
    assert 2 == len(sample_mlst_alleles)

    with pytest.raises(EntityExistsError) as execinfo:
        mlst_service.insert(feature_scope_name='lmonocytogenes', data_package=mlst_data_package_single_scheme)

    assert 'Passed samples already have features for feature scope [lmonocytogenes]' in str(execinfo.value)

    sample_mlst_alleles = session.query(SampleMLSTAlleles).all()
    assert 2 == len(sample_mlst_alleles)


def test_multiple_insert_different_samples_same_scheme(database, mlst_data_package_single_scheme,
                                                       mlst_data_package_single_scheme2,
                                                       sample_service, filesystem_storage):
    mlst_service = MLSTService(database_connection=database,
                               sample_service=sample_service,
                               mlst_dir=filesystem_storage.mlst_dir)

    session = database.get_session()

    # Insert first set of data
    mlst_service.insert(feature_scope_name='lmonocytogenes', data_package=mlst_data_package_single_scheme)

    features_count = session.query(MLSTAllelesSamples).count()
    assert 7 == features_count

    sample_CFSAN002349 = session.query(Sample).filter(Sample.name == 'CFSAN002349').one()
    sample_CFSAN023463 = session.query(Sample).filter(Sample.name == 'CFSAN023463').one()

    ## Test feature where all new samples get added
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'cat',
        'allele': '11',
    })
    assert m.id == 'lmonocytogenes:cat:11'
    assert m.sla == 'lmonocytogenes:cat:11'
    assert set(m.sample_ids) == {sample_CFSAN002349.id, sample_CFSAN023463.id}

    ## Test feature where only one of the new samples gets added
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'abcZ',
        'allele': '1',
    })
    assert m.id == 'lmonocytogenes:abcZ:1'
    assert m.sla == 'lmonocytogenes:abcZ:1'
    assert set(m.sample_ids) == {sample_CFSAN002349.id, sample_CFSAN023463.id}

    ## Test feature where none of the new samples gets added
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'dapE',
        'allele': '13',
    })
    assert m.id == 'lmonocytogenes:dapE:13'
    assert m.sla == 'lmonocytogenes:dapE:13'
    assert set(m.sample_ids) == {sample_CFSAN002349.id, sample_CFSAN023463.id}

    ## Test feature that only exists after the 2nd addition
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'dapE',
        'allele': '12',
    })
    assert m is None

    # Insert second set of data
    mlst_service.insert(feature_scope_name='lmonocytogenes', data_package=mlst_data_package_single_scheme2)

    features_count = session.query(MLSTAllelesSamples).count()
    assert 12 == features_count

    sample_CFSAN002349_2 = session.query(Sample).filter(Sample.name == 'CFSAN002349-2').one()
    sample_CFSAN023463_2 = session.query(Sample).filter(Sample.name == 'CFSAN023463-2').one()

    ## Test feature where all new samples get added
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'cat',
        'allele': '11',
    })
    assert m.id == 'lmonocytogenes:cat:11'
    assert m.sla == 'lmonocytogenes:cat:11'
    assert set(m.sample_ids) == {sample_CFSAN002349.id, sample_CFSAN023463.id,
                                 sample_CFSAN002349_2.id, sample_CFSAN023463_2.id}

    ## Test feature where only one of the new samples gets added
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'abcZ',
        'allele': '1',
    })
    assert m.id == 'lmonocytogenes:abcZ:1'
    assert m.sla == 'lmonocytogenes:abcZ:1'
    assert set(m.sample_ids) == {sample_CFSAN002349.id, sample_CFSAN023463.id,
                                 sample_CFSAN002349_2.id}

    ## Test feature where none of the new samples gets added
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'dapE',
        'allele': '13',
    })
    assert m.id == 'lmonocytogenes:dapE:13'
    assert m.sla == 'lmonocytogenes:dapE:13'
    assert set(m.sample_ids) == {sample_CFSAN002349.id, sample_CFSAN023463.id}

    ## Test feature that only exists after the 2nd addition
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'dapE',
        'allele': '12',
    })
    assert m.id == 'lmonocytogenes:dapE:12'
    assert m.sla == 'lmonocytogenes:dapE:12'
    assert set(m.sample_ids) == {sample_CFSAN002349_2.id, sample_CFSAN023463_2.id}


def test_multiple_inserts_different_schemes(database, mlst_data_package_basic,
                                            mlst_data_package_single_scheme2,
                                            sample_service, filesystem_storage):
    mlst_service = MLSTService(database_connection=database,
                               sample_service=sample_service,
                               mlst_dir=filesystem_storage.mlst_dir)

    session = database.get_session()

    mlst_service.insert(data_package=mlst_data_package_single_scheme2)
    mlst_service.insert(data_package=mlst_data_package_basic)

    features_count = session.query(MLSTAllelesSamples).count()
    assert 28 == features_count

    sample_2014C_3598 = session.query(Sample).filter(Sample.name == '2014C-3598').one()
    sample_2014C_3599 = session.query(Sample).filter(Sample.name == '2014C-3599').one()
    sample_CFSAN002349_2 = session.query(Sample).filter(Sample.name == 'CFSAN002349-2').one()
    sample_CFSAN023463_2 = session.query(Sample).filter(Sample.name == 'CFSAN023463-2').one()
    sample_CFSAN002349 = session.query(Sample).filter(Sample.name == 'CFSAN002349').one()
    sample_CFSAN023463 = session.query(Sample).filter(Sample.name == 'CFSAN023463').one()

    # Test feature where only some of the new samples gets added
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'abcZ',
        'allele': '1',
    })
    assert m.id == 'lmonocytogenes:abcZ:1'
    assert m.sla == 'lmonocytogenes:abcZ:1'
    assert set(m.sample_ids) == {sample_CFSAN002349.id, sample_CFSAN023463.id,
                                 sample_CFSAN002349_2.id}

    # Test feature with unknown alleles
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'lhkA',
        'allele': '?',
    })
    assert m.id == 'lmonocytogenes:lhkA:?'
    assert m.sla == 'lmonocytogenes:lhkA:?'
    assert set(m.sample_ids) == {sample_CFSAN023463_2.id}

    # Test completely different scheme
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'ecoli',
        'locus': 'adk',
        'allele': '100',
    })
    assert m.id == 'ecoli:adk:100'
    assert m.sla == 'ecoli:adk:100'
    assert set(m.sample_ids) == {sample_2014C_3598.id, sample_2014C_3599.id}


def test_multiple_inserts_3_inserts(database, mlst_data_package_single_scheme,
                                    mlst_data_package_single_scheme2,
                                    mlst_data_package_single_scheme3,
                                    sample_service, filesystem_storage):
    mlst_service = MLSTService(database_connection=database,
                               sample_service=sample_service,
                               mlst_dir=filesystem_storage.mlst_dir)

    session = database.get_session()

    mlst_service.insert(feature_scope_name='lmonocytogenes', data_package=mlst_data_package_single_scheme)
    mlst_service.insert(feature_scope_name='lmonocytogenes', data_package=mlst_data_package_single_scheme2)
    mlst_service.insert(feature_scope_name='lmonocytogenes', data_package=mlst_data_package_single_scheme3)

    features_count = session.query(MLSTAllelesSamples).count()
    assert 13 == features_count

    sample_CFSAN002349 = session.query(Sample).filter(Sample.name == 'CFSAN002349').one()
    sample_CFSAN023463 = session.query(Sample).filter(Sample.name == 'CFSAN023463').one()
    sample_CFSAN002349_2 = session.query(Sample).filter(Sample.name == 'CFSAN002349-2').one()
    sample_CFSAN023463_2 = session.query(Sample).filter(Sample.name == 'CFSAN023463-2').one()
    sample_CFSAN002349_3 = session.query(Sample).filter(Sample.name == 'CFSAN002349-3').one()
    sample_CFSAN023463_3 = session.query(Sample).filter(Sample.name == 'CFSAN023463-3').one()

    # Test feature where only some of the new samples gets added
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'abcZ',
        'allele': '1',
    })
    assert m.id == 'lmonocytogenes:abcZ:1'
    assert m.sla == 'lmonocytogenes:abcZ:1'
    assert set(m.sample_ids) == {sample_CFSAN002349.id, sample_CFSAN023463.id,
                                 sample_CFSAN002349_2.id,
                                 sample_CFSAN002349_3.id}

    # Test feature with unknown alleles
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'lhkA',
        'allele': '?',
    })
    assert m.id == 'lmonocytogenes:lhkA:?'
    assert m.sla == 'lmonocytogenes:lhkA:?'
    assert set(m.sample_ids) == {sample_CFSAN023463_2.id,
                                 sample_CFSAN023463_3.id}

    # Test valid alleles for this same locus
    m: MLSTAllelesSamples = session.query(MLSTAllelesSamples).get({
        'scheme': 'lmonocytogenes',
        'locus': 'lhkA',
        'allele': '5',
    })
    assert m.id == 'lmonocytogenes:lhkA:5'
    assert m.sla == 'lmonocytogenes:lhkA:5'
    assert set(m.sample_ids) == {sample_CFSAN002349.id, sample_CFSAN023463.id,
                                 sample_CFSAN002349_2.id,
                                 sample_CFSAN002349_3.id}


def test_get_mlst_schemes(mlst_service_loaded):
    mlst_schemes = mlst_service_loaded.get_mlst_schemes()
    assert 3 == len(mlst_schemes)
    assert isinstance(mlst_schemes[0], MLSTScheme)
    assert {'lmonocytogenes', 'ecoli', 'campylobacter'} == {s.name for s in mlst_schemes}


def test_get_all_mlst_features(mlst_service_loaded: MLSTService):
    # Test no unknown
    mlst_features = mlst_service_loaded.get_features(include_present=True, include_unknown=False)
    assert 22 == len(mlst_features)
    assert {'lmonocytogenes:abcZ:1', 'lmonocytogenes:bglA:51', 'lmonocytogenes:cat:11',
            'lmonocytogenes:dapE:13', 'lmonocytogenes:dat:2', 'lmonocytogenes:ldh:5',
            'lmonocytogenes:lhkA:4', 'lmonocytogenes:lhkA:5',
            'ecoli:adk:100', 'ecoli:fumC:23', 'ecoli:gyrB:68', 'ecoli:icd:45', 'ecoli:mdh:1',
            'ecoli:purA:35', 'ecoli:recA:7',
            'campylobacter:aspA:2', 'campylobacter:glnA:1', 'campylobacter:gltA:1',
            'campylobacter:glyA:3', 'campylobacter:pgm:2', 'campylobacter:tkt:1',
            'campylobacter:uncA:6'} == set(mlst_features.keys())
    assert isinstance(mlst_features['lmonocytogenes:abcZ:1'], MLSTAllelesSamples)

    # Test include unknown
    mlst_features = mlst_service_loaded.get_features(include_present=True, include_unknown=True)
    assert 23 == len(mlst_features)
    assert {'lmonocytogenes:abcZ:1', 'lmonocytogenes:bglA:51', 'lmonocytogenes:cat:11',
            'lmonocytogenes:dapE:13', 'lmonocytogenes:dat:2', 'lmonocytogenes:ldh:5',
            'lmonocytogenes:lhkA:4', 'lmonocytogenes:lhkA:5',
            'ecoli:adk:100', 'ecoli:fumC:23', 'ecoli:gyrB:68', 'ecoli:icd:45', 'ecoli:mdh:1',
            'ecoli:purA:35', 'ecoli:recA:7',
            'campylobacter:aspA:2', 'campylobacter:glnA:1', 'campylobacter:gltA:1',
            'campylobacter:glyA:3', 'campylobacter:pgm:2', 'campylobacter:tkt:1',
            'campylobacter:uncA:6', 'campylobacter:uncA:?'} == set(mlst_features.keys())
    assert isinstance(mlst_features['lmonocytogenes:abcZ:1'], MLSTAllelesSamples)

    # Test only unknown
    mlst_features = mlst_service_loaded.get_features(include_present=False, include_unknown=True)
    assert 1 == len(mlst_features)
    assert {'campylobacter:uncA:?'} == set(mlst_features.keys())
    assert isinstance(mlst_features['campylobacter:uncA:?'], MLSTAllelesSamples)


def test_get_features_for_scheme(mlst_service_loaded: MLSTService):
    # Test lmonocytogenes
    mlst_features = mlst_service_loaded.get_features('lmonocytogenes')
    assert 8 == len(mlst_features)
    assert {'lmonocytogenes:abcZ:1', 'lmonocytogenes:bglA:51', 'lmonocytogenes:cat:11',
            'lmonocytogenes:dapE:13', 'lmonocytogenes:dat:2', 'lmonocytogenes:ldh:5',
            'lmonocytogenes:lhkA:4', 'lmonocytogenes:lhkA:5'} == set(mlst_features.keys())
    assert isinstance(mlst_features['lmonocytogenes:abcZ:1'], MLSTAllelesSamples)

    # Test campylobacter
    mlst_features = mlst_service_loaded.get_features('campylobacter')
    assert 7 == len(mlst_features)
    assert {'campylobacter:aspA:2', 'campylobacter:glnA:1', 'campylobacter:gltA:1',
            'campylobacter:glyA:3', 'campylobacter:pgm:2', 'campylobacter:tkt:1',
            'campylobacter:uncA:6'} == set(mlst_features.keys())
    assert isinstance(mlst_features['campylobacter:glyA:3'], MLSTAllelesSamples)

    # Test lmonocytogenes with specific locus
    mlst_features = mlst_service_loaded.get_features('lmonocytogenes', locus='abcZ')
    assert 1 == len(mlst_features)
    assert {'lmonocytogenes:abcZ:1'} == set(mlst_features.keys())

    # Test lmonocytogenes with specific locus 2
    mlst_features = mlst_service_loaded.get_features('lmonocytogenes', locus='lhkA')
    assert 2 == len(mlst_features)
    assert {'lmonocytogenes:lhkA:4', 'lmonocytogenes:lhkA:5'} == set(mlst_features.keys())

    # Test include unknown
    mlst_features = mlst_service_loaded.get_features('campylobacter', include_unknown=True)
    assert 8 == len(mlst_features)
    assert {'campylobacter:aspA:2', 'campylobacter:glnA:1', 'campylobacter:gltA:1',
            'campylobacter:glyA:3', 'campylobacter:pgm:2', 'campylobacter:tkt:1',
            'campylobacter:uncA:6', 'campylobacter:uncA:?'} == set(mlst_features.keys())
    assert isinstance(mlst_features['campylobacter:glyA:3'], MLSTAllelesSamples)

    # Test include only unknown (not present)
    mlst_features = mlst_service_loaded.get_features('campylobacter', include_present=False,
                                                     include_unknown=True)
    assert 1 == len(mlst_features)
    assert {'campylobacter:uncA:?'} == set(mlst_features.keys())
    assert isinstance(mlst_features['campylobacter:uncA:?'], MLSTAllelesSamples)

    # Test include neither present nor unknown
    mlst_features = mlst_service_loaded.get_features('campylobacter', include_present=False,
                                                     include_unknown=False)
    assert 0 == len(mlst_features)

    # Test invalid scheme
    assert 0 == len(mlst_service_loaded.get_features('invalid_scheme'))


def test_get_all_alleles(mlst_service_loaded: MLSTService):
    assert {'1'} == mlst_service_loaded.get_all_alleles('lmonocytogenes', 'abcZ')


def test_get_all_alleles_multiple_alleles(mlst_service_loaded: MLSTService):
    assert {'4', '5'} == mlst_service_loaded.get_all_alleles('lmonocytogenes', 'lhkA')


def test_get_all_alleles_unknown_alleles(mlst_service_loaded_unknown: MLSTService):
    assert {'1', '?'} == mlst_service_loaded_unknown.get_all_alleles('lmonocytogenes', 'abcZ')


def test_get_all_alleles_unknown_alleles2(mlst_service_loaded_unknown: MLSTService):
    assert {'?'} == mlst_service_loaded_unknown.get_all_alleles('lmonocytogenes', 'bglA')


def test_get_all_loci_alleles(mlst_service_loaded: MLSTService):
    assert {('abcZ', '1'), ('bglA', '51'), ('cat', '11'),
            ('dapE', '13'), ('dat', '2'), ('ldh', '5'),
            ('lhkA', '4'), ('lhkA', '5')} == mlst_service_loaded.get_all_loci_alleles('lmonocytogenes')


def test_get_all_loci_alleles_unknown(mlst_service_loaded_unknown: MLSTService):
    assert {('abcZ', '1'), ('abcZ', '?'), ('bglA', '?'), ('cat', '11'),
            ('dapE', '13'), ('dat', '2'), ('ldh', '5'),
            ('lhkA', '5')} == mlst_service_loaded_unknown.get_all_loci_alleles('lmonocytogenes')



from __future__ import annotations

import bz2
import gzip
import logging
import lzma
from functools import partial
from mimetypes import guess_type
from os import mkdir, symlink, path
from os.path import basename, splitext
from pathlib import Path
from typing import Tuple, List, Generator

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from jinja2 import Environment, FileSystemLoader, select_autoescape

from genomics_data_index.storage.io.mutation.SnpEffDatabase import SnpEffDatabase
from genomics_data_index.storage.util import execute_commands, TRACE_LEVEL

logger = logging.getLogger(__name__)

template_dir = Path(path.dirname(__file__)) / 'templates'

jinja_env = Environment(
    loader=FileSystemLoader(template_dir),
    autoescape=select_autoescape()
)


class SequenceFile:

    def __init__(self, file: Path):
        self._file = file

    @property
    def file(self) -> Path:
        return self._file

    def parse_sequence_file(self) -> Tuple[str, List[SeqRecord]]:
        _open = self._get_open_partial()
        ref_name = self.get_genome_name()

        with _open(self._file) as f:
            if self.is_genbank():
                logger.log(TRACE_LEVEL, f'Parsing sequence file [{self._file}] as a genbank file')
                sequences = list(SeqIO.parse(f, 'genbank'))
            else:
                logger.log(TRACE_LEVEL, f'Parsing sequence file [{self._file}] as a fasta file')
                sequences = list(SeqIO.parse(f, 'fasta'))

            return ref_name, sequences

    def _get_open_partial(self):
        # Code for handling gzipped/non-gzipped derived https://stackoverflow.com/a/52839332
        if self.is_gzip():
            return partial(gzip.open, mode='rt')
        elif self.is_lzma():
            return partial(lzma.open, mode='rt')
        elif self.is_bzip2():
            return partial(bz2.open, mode='rt')
        else:
            return open

    def records(self) -> Generator[SeqRecord, None, None]:
        _open = self._get_open_partial()

        if self.is_fasta():
            format = 'fasta'
        elif self.is_genbank():
            format = 'genbank'
        else:
            raise Exception(f'sequence_file={self.file} is neither genbank nor fasta')

        with _open(self._file) as f:
            logger.debug(f'Parsing sequence file [{self._file}]')
            yield from SeqIO.parse(f, format)
        logger.debug(f'Finished parsing sequence file [{self._file}]')

    def write(self, output_file: Path, format: str) -> None:
        ref_name, sequences = self.parse_sequence_file()
        with open(output_file, 'w') as fh:
            SeqIO.write(sequences, fh, format)

    def get_genome_extension_minus_compression(self):
        if self.is_gzip():
            ref_extension = splitext(basename(self._file).rstrip('.gz'))[1]
        elif self.is_lzma():
            ref_extension = splitext(basename(self._file).rstrip('.xz'))[1]
        elif self.is_bzip2():
            ref_extension = splitext(basename(self._file).rstrip('.bz2'))[1]
        else:
            ref_extension = splitext(basename(self._file))[1]

        logger.log(TRACE_LEVEL, f'Sequence file [{self._file}] has extension (minus compression) [{ref_extension}]')

        return ref_extension

    def is_gzip(self) -> bool:
        encoding = guess_type(str(self._file))[1]  # uses file extension
        return encoding == 'gzip'

    def is_lzma(self) -> bool:
        encoding = guess_type(str(self._file))[1]  # uses file extension
        return encoding == 'xz'

    def is_bzip2(self) -> bool:
        encoding = guess_type(str(self._file))[1]  # uses file extension
        return encoding == 'bzip2'

    def is_genbank(self) -> bool:
        ref_extension = self.get_genome_extension_minus_compression().lower()
        return ref_extension == '.gb' or ref_extension == '.gbk'

    def is_fasta(self) -> bool:
        ref_extension = self.get_genome_extension_minus_compression().lower()
        return ref_extension == '.fasta' or ref_extension == '.fa' or ref_extension == '.fna'

    def is_assembly(self) -> bool:
        return self.is_fasta() or self.is_genbank()

    def is_reads(self) -> bool:
        extension = self.get_genome_extension_minus_compression().lower()
        return extension == '.fastq' or extension == '.fq'

    def name_differences(self, other_file: SequenceFile) -> List[str]:
        """
        Gets the differences from this sequence file name and the passed file name.
        I assume both file names are of the same length (i.e., no insertion of gaps anywhere or
        complicated alignment of strings). This is intended to be used to figure out which of a pair
        of fastq files is the forward and which is the reverse.
        :param other_file: The other sequence file.
        :return: A list of strings containing mismatches between this file name and the other file name.
        """
        name1 = self._file.name
        name2 = other_file._file.name

        if len(name1) != len(name2):
            raise Exception(f'name1=[{name1}] is not the same length as name2=[{name2}]')
        else:
            differences = []
            current_difference = None
            previous_difference_index = -1
            for i in range(len(name1)):
                c1 = name1[i]
                c2 = name2[i]
                if c1 == c2:
                    if current_difference is not None:
                        differences.append(current_difference)
                        current_difference = None
                else:
                    if previous_difference_index == (i - 1):
                        if current_difference is None:
                            current_difference = c1
                            previous_difference_index = i
                        else:
                            current_difference += c1
                            previous_difference_index = i
                    elif current_difference is None:
                        current_difference = c1
                        previous_difference_index = i
                    else:
                        raise Exception(f'previous_difference_index=[{previous_difference_index}], i=[{i}], '
                                        f'but current_difference is not None')

            # Append last bit of differences
            if current_difference is not None:
                differences.append(current_difference)

            return differences

    def can_use_snpeff(self) -> bool:
        return self.is_genbank()

    def get_genome_name(self, exclude_paired_end_indicators: bool = False) -> str:
        """
        Gets the genome name (filename minus extension). Accounts for compressed files.
        :param file: The file.
        :param exclude_paired_end_indicators: Exclude paired-end indicators when getting genome name (e.g., _R1, _1, etc).
                                              this only applies if the sequence file is a reads file.
        :return: The genome name from the file.
        """
        if self.is_gzip():
            name_minus_extension = splitext(basename(self._file).rstrip('.gz'))[0]
        elif self.is_lzma():
            name_minus_extension = splitext(basename(self._file).rstrip('.xz'))[0]
        elif self.is_bzip2():
            name_minus_extension = splitext(basename(self._file).rstrip('.bz2'))[0]
        else:
            name_minus_extension = splitext(basename(self._file))[0]

        if self.is_reads() and exclude_paired_end_indicators:
            name_minus_pair_indicator = name_minus_extension
            for pair_indicator in ['_1', '_2', '_R1', '_R2', '_f', '_r']:
                last_index = name_minus_extension.rfind(pair_indicator)
                if last_index != -1:
                    name_minus_pair_indicator = name_minus_extension[:last_index]
                    break
            return name_minus_pair_indicator
        else:
            return name_minus_extension

    def _write_snpeff_config(self, snpeff_config_file: Path, snpeff_database_dir: Path,
                             codon_type: str) -> str:
        genome_name, records = self.parse_sequence_file()
        sequence_ids = [s.id for s in records]

        snpeff_config_template = jinja_env.get_template('snpEff.config')
        with open(snpeff_config_file, 'w') as snpeff_config:
            snpeff_config.write(snpeff_config_template.render(
                data_dir=str(snpeff_database_dir),
                reference_name=genome_name,
                sequence_ids=sequence_ids,
                codon_type=codon_type,
            ))

        logger.debug(f'Wrote snpeff config file [{snpeff_config_file}]')

        return genome_name

    def _setup_snpeff_files(self, snpeff_database_dir: Path, reference_name: str):
        if not snpeff_database_dir.exists():
            mkdir(snpeff_database_dir)

        reference_dir = snpeff_database_dir / reference_name
        mkdir(reference_dir)

        if self.is_gzip():
            genbank_path = reference_dir / 'genes.gbk.gz'
        else:
            genbank_path = reference_dir / 'genes.gbk'
        symlink(self._file, genbank_path)

    def create_snpeff_database(self, database_dir: Path, codon_type: str = 'Standard') -> SnpEffDatabase:
        if not self.is_genbank():
            raise Exception(f'Sequence file [{self._file}] is not a genbank file. '
                            f'Can only build snpeff databases for genbank files.')

        logger.debug(f'Setting up snpeff database in [{database_dir}]')

        snpeff_config_path = database_dir / 'snpEff.config'
        snpeff_database_dir = database_dir / 'db'

        genome_name = self._write_snpeff_config(
            snpeff_config_file=snpeff_config_path,
            snpeff_database_dir=snpeff_database_dir,
            codon_type=codon_type)

        self._setup_snpeff_files(snpeff_database_dir=snpeff_database_dir,
                                 reference_name=genome_name)

        command = ['snpEff', 'build', '-v', '-c', str(snpeff_config_path), '-genbank', genome_name]
        execute_commands([command])

        return SnpEffDatabase(snpeff_config=snpeff_config_path,
                              database_dir=database_dir,
                              genome_name=genome_name)



import logging
from typing import List, Dict, Optional

import pandas as pd

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.io.SampleData import SampleData
from genomics_data_index.storage.io.mutation.NucleotideFeaturesReader import NucleotideFeaturesReader
from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData
from genomics_data_index.storage.io.mutation.VcfSnpEffAnnotationParser import VcfSnpEffAnnotationParser
from genomics_data_index.storage.io.mutation.variants_processor.VcfVariantsTableProcessor import \
    VcfVariantsTableProcessorFactory
from genomics_data_index.storage.util.SamplesProgressLogger import SamplesProgressLogger

logger = logging.getLogger(__name__)


class VcfVariantsReader(NucleotideFeaturesReader):
    VCF_FRAME_COLUMNS = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID']

    def __init__(self, sample_files_map: Dict[str, NucleotideSampleData],
                 variants_processor_factory: VcfVariantsTableProcessorFactory,
                 progress_logger: SamplesProgressLogger = None,
                 include_masked_regions: bool = True):
        super().__init__()
        self._sample_files_map = sample_files_map
        self._snpeff_parser = VcfSnpEffAnnotationParser()
        self._include_masked_regions = include_masked_regions
        self._variants_processor_factory = variants_processor_factory
        self._progress_logger = progress_logger

    def get_or_create_feature_file(self, sample_name: str):
        vcf_file, index_file = self._sample_files_map[sample_name].get_vcf_file()
        return vcf_file

    def _read_features_table(self) -> pd.DataFrame:
        frames = []
        sample_data_list = list(self._sample_files_map.values())
        num_samples = len(sample_data_list)
        logger.debug(f'Starting to read features table from {num_samples} sample files')
        variants_processor = self._variants_processor_factory.create_variants_processor(self._include_masked_regions)
        samples_processed = 0
        progress_logger = self._progress_logger
        if progress_logger is None:
            progress_logger = SamplesProgressLogger(stage_name='Features', stage_number=-1, total_samples=num_samples)
        update_every_nth_sample = progress_logger.get_update_every_nth_sample(percent_to_update=2)

        progress_logger.update_progress(0)
        for frame_vcf_mask in variants_processor.process(sample_data_list):
            frames.append(frame_vcf_mask)
            if (samples_processed % update_every_nth_sample) == 0:
                progress_logger.update_progress(samples_processed)
            samples_processed = samples_processed + 1
        progress_logger.update_progress(samples_processed)

        logger.debug(f'Finished reading features table from {num_samples} samples')
        return pd.concat(frames)

    def get_sample_files(self, sample_name: str) -> Optional[SampleData]:
        return self._sample_files_map[sample_name]

    def get_genomic_masked_region(self, sample_name: str) -> MaskedGenomicRegions:
        return self._sample_files_map[sample_name].get_mask()

    def samples_list(self) -> List[str]:
        return list(self._sample_files_map.keys())

    @classmethod
    def create(cls, sample_files_map: Dict[str, NucleotideSampleData],
               variants_processor_factory: VcfVariantsTableProcessorFactory,
               include_masked_regions: bool = True):
        return cls(sample_files_map=sample_files_map,
                   variants_processor_factory=variants_processor_factory,
                   include_masked_regions=include_masked_regions)
